import numpy as np

# Originally developed for MATLAB by Benjamin Schafer PhD et al
# Ported to Python by Brooks Smith MEng, PE, CPEng
#
# Each function within this file was originally its own separate file.
# Original MATLAB comments, especially those retaining to authorship or
# change history, have been generally retained unaltered


def m_sort(m_all):
    for i, m_a in enumerate(m_all):
        # return all the nonzeros longitudinal terms in m_a as a column vector
        m_a = m_a[np.nonzero(m_a)]
        m_a = np.lib.arraysetops.unique(m_a)  # remove repetitive longitudinal terms
        m_all[i] = m_a
    return m_all


def constr_bc_flag(nodes, constraints):
    # this subroutine is to determine flags for user constraints and internal (at node) B.C.'s

    # inputs:
    # nodes: [node# x y dof_x dof_y dof_z dof_r stress] n_nodes x 8
    # constraints:: [node# e dof_e coeff node# k dof_k] e=dof to be eliminated
    #   k=kept dof dofe_node = coeff*dofk_nodek

    # Output:
    # bc_flag: 1 if there are user constraints or node fixities
    # 0 if there is no user constraints and node fixities

    # Z. Li, June 2010

    # Check for boundary conditions on the nodes
    n_nodes = len(nodes)
    for i in range(0, n_nodes):
        for j in range(3, 7):
            if nodes[i, j] == 0:
                return 1

    # Check for user defined constraints too
    if len(constraints) == 0:
        return 0
    else:
        return 1


def addspring(k_global, springs, n_nodes, length, b_c, m_a):
    # BWS
    # August 2000

    # [k_global]=addspring(k_global,springs,n_nodes,length,m_a,b_c)
    # Add spring stiffness to global elastic stiffness matrix

    # k_global is the complete elastic stiffness matrix
    # springs is the definition of any external springs added to the member
    # springs=[node# DOF(x=1,y=2,z=3,theta=4) k_s]

    # modified by Z. Li, Aug. 09, 2009 for general B.C.
    # Z. Li, June 2010

    if len(springs) > 0:
        total_m = len(m_a)  # Total number of longitudinal terms m

        for spring in springs:
            node = spring[0]
            dof = spring[1]
            k_stiff = spring[2]
            k_flag = spring[3]

            if dof == 1:
                r_c = 2*node - 1
            elif dof == 2:
                r_c = 2*n_nodes + 2*node - 1
            elif dof == 3:
                r_c = 2 * node
            elif dof == 4:
                r_c = 2*n_nodes + 2*node
            else:
                r_c = 1
                k_s = 0

            for i in range(0, total_m):
                for j in range(0, total_m):
                    if k_flag == 0:
                        k_s = k_stiff  # k_stiff is the total stiffness and may be added directly
                    else:
                        if dof == 3:
                            k_s = 0  # axial dof with a foundation stiffness has no net stiffness
                        else:
                            [i_1, _, _, _, _] = bc_i1_5(b_c, m_a[i], m_a[j], length)
                            k_s = k_stiff * i_1
                            # k_stiff is a foundation stiffness and an equivalent
                            # total stiffness must be calculated

                    k_global[4*n_nodes*i+r_c-1, 4*n_nodes*j+r_c-1] \
                        = k_global[4*n_nodes*i+r_c-1, 4*n_nodes*j+r_c-1] + k_s

    return k_global


def elem_prop(nodes, elements):
    el_props = np.zeros((len(elements), 3))
    for i, elem in enumerate(elements):
        node_i = int(elem[1])
        node_j = int(elem[2])
        x_i = nodes[node_i, 1]
        y_i = nodes[node_i, 2]
        x_j = nodes[node_j, 1]
        y_j = nodes[node_j, 2]
        d_x = x_j - x_i
        d_y = y_j - y_i
        width = np.sqrt(d_x**2 + d_y**2)
        alpha = np.arctan2(d_y, d_x)
        el_props[i] = [i, width, alpha]
    return el_props


def k_kg_global(nodes, elements, el_props, props, length, b_c, m_a):
    total_m = len(m_a)
    n_nodes = len(nodes)
    # ZERO OUT THE GLOBAL MATRICES
    k_global = np.zeros((4 * n_nodes * total_m, 4 * n_nodes * total_m))
    kg_global = np.zeros((4 * n_nodes * total_m, 4 * n_nodes * total_m))

    # ASSEMBLE THE GLOBAL STIFFNESS MATRICES
    for i, elem in enumerate(elements):
        # Generate element stiffness matrix (k_local) in local coordinates
        # Generate geometric stiffness matrix (kg_local) in local coordinates
        thick = elem[3]
        b_strip = el_props[i, 1]
        mat_num = int(elem[4])
        row = int((np.argwhere(props[:, 0] == mat_num)).reshape(1))
        mat = props[row]
        stiff_x = mat[1]
        stiff_y = mat[2]
        nu_x = mat[3]
        nu_y = mat[4]
        bulk = mat[5]

        node_i = int(elem[1])
        node_j = int(elem[2])
        ty_1 = nodes[node_i][7] * thick
        ty_2 = nodes[node_j][7] * thick

        k_l, kg_l = k_kg_local(
            stiff_x=stiff_x,
            stiff_y=stiff_y,
            nu_x=nu_x,
            nu_y=nu_y,
            bulk=bulk,
            thick=thick,
            length=length,
            ty_1=ty_1,
            ty_2=ty_2,
            b_strip=b_strip,
            b_c=b_c,
            m_a=m_a
        )

        # Transform k_local and kg_local into global coordinates
        alpha = el_props[i, 2]
        gamma = trans(alpha=alpha, total_m=total_m)
        k_local = gamma @ k_l @ gamma.conj().T
        kg_local = gamma @ kg_l @ gamma.conj().T

        # Add element contribution of k_local to full matrix k_global and kg_local to kg_global
        k_global, kg_global = assemble(
            k_global=k_global,
            kg_global=kg_global,
            k_local=k_local,
            kg_local=kg_local,
            node_i=node_i,
            node_j=node_j,
            n_nodes=n_nodes,
            m_a=m_a
        )

    return k_global, kg_global


def k_kg_local(stiff_x, stiff_y, nu_x, nu_y, bulk, thick, length, ty_1, ty_2, b_strip, b_c, m_a):
    # Generate element stiffness matrix (k_local) in local coordinates
    # Generate geometric stiffness matrix (kg_local) in local coordinates

    # Inputs:
    # stiff_x,stiff_y,nu_x,nu_y,bulk: material properties
    # thick: thickness of the strip (element)
    # length: length of the strip in longitudinal direction
    # ty_1, ty_2: node stresses
    # b_strip: width of the strip in transverse direction
    # b_c: ['S-S'] a string specifying boundary conditions to be analyzed:
    # 'S-S' simply-pimply supported boundary condition at loaded edges
    # 'C-C' clamped-clamped boundary condition at loaded edges
    # 'S-C' simply-clamped supported boundary condition at loaded edges
    # 'C-F' clamped-free supported boundary condition at loaded edges
    # 'C-G' clamped-gcdef np.ndarray[np.double_t, ndim=2] uided supported boundary condition at
    #       loaded edges
    # m_a: longitudinal terms (or half-wave numbers) for this length

    # Output:
    # k_local: local stiffness matrix, a total_m x total_m matrix of 8 by 8 submatrices.
    # k_local=[k_mp]total_m x total_m block matrix
    # each k_mp is the 8 x 8 submatrix in the DOF order [u1 v1 u2 v2 w1 theta1 w2 theta2]'
    # kg_local: local geometric stiffness matrix, a total_m x total_m matrix of 8 by 8 submatrices.
    # kg_local=[kg_mp]total_m x total_m block matrix
    # each kg_mp is the 8 x 8 submatrix in the DOF order [u1 v1 u2 v2 w1 theta1
    # w2 theta2]'

    # Z. Li June 2008
    # modified by Z. Li, Aug. 09, 2009
    # modified by Z. Li, June 2010
    # klocal and kglocal merged by B Smith, May 2023

    e_1 = stiff_x / (1 - nu_x*nu_y)
    e_2 = stiff_y / (1 - nu_x*nu_y)
    d_x = stiff_x * thick**3 / (12 * (1 - nu_x*nu_y))
    d_y = stiff_y * thick**3 / (12 * (1 - nu_x*nu_y))
    d_1 = nu_x * stiff_y * thick**3 / (12 * (1 - nu_x*nu_y))
    d_xy = bulk * thick**3 / 12

    total_m = len(m_a)  # Total number of longitudinal terms m

    k_local = np.zeros((8 * total_m, 8 * total_m))
    kg_local = np.zeros((8 * total_m, 8 * total_m))

    for i in range(0, total_m):
        for j in range(0, total_m):
            u_i = m_a[i] * np.pi
            u_j = m_a[j] * np.pi
            c_1 = u_i / length
            c_2 = u_j / length

            [i_1, i_2, i_3, i_4, i_5] = bc_i1_5(b_c=b_c, m_i=m_a[i], m_j=m_a[j], length=length)

            k_local[8 * i:8*i + 4, 8 * j:8*j + 4] = calc_km_mp(
                e_1=e_1,
                e_2=e_2,
                c_1=c_1,
                c_2=c_2,
                b_strip=b_strip,
                bulk=bulk,
                nu_x=nu_x,
                thick=thick,
                i_1=i_1,
                i_2=i_2,
                i_3=i_3,
                i_4=i_4,
                i_5=i_5
            )
            k_local[8*i + 4:8 * (i+1), 8*j + 4:8 * (j+1)] = calc_kf_mp(
                d_x=d_x,
                d_y=d_y,
                d_1=d_1,
                d_xy=d_xy,
                b_strip=b_strip,
                i_1=i_1,
                i_2=i_2,
                i_3=i_3,
                i_4=i_4,
                i_5=i_5
            )

            kg_local[8 * i:8*i + 4, 8 * j:8*j + 4] = calc_gm_mp(
                u_i=u_i,
                u_j=u_j,
                b_strip=b_strip,
                length=length,
                ty_1=ty_1,
                ty_2=ty_2,
                i_4=i_4,
                i_5=i_5
            )
            kg_local[8*i + 4:8 * (i+1), 8*j + 4:8 * (j+1)] = calc_gf_mp(
                ty_1=ty_1, ty_2=ty_2, b_strip=b_strip, i_5=i_5
            )

    return k_local, kg_local


def kglobal_transv(nodes, elements, props, m_i, length, b_c, el_props):
    #
    # this routine creates the global stiffness matrix for planar displacements
    # basically the same way as in the main program, however:
    #   only one half-wave number m_i is considered,
    #   only w, teta terms are considered,
    #   plus stiff_y = nu_x = nu_y = 0 is assumed
    #   plus the longitudinal displacements. DOFs are explicitely eliminated
    #   the multiplication by 'length' (member length) is not done here, must be done
    #      outside of this routine
    #
    # input/output data
    #   nodes, elements, props - same as elsewhere throughout this program
    #   m_i - number of half waves
    #   k_global_transv - global stiffness matrix (geometric not included)
    #
    # S. Adany, Feb 08, 2004
    # Z. Li, Jul 10, 2009
    #
    n_nodes = len(nodes)
    k_global_transv = np.zeros((4 * n_nodes, 4 * n_nodes))

    for i, elem in enumerate(elements):
        thick = elem[3]
        b_strip = el_props[i, 1]
        mat_num = int(elem[4])
        row = int((np.argwhere(props[:, 0] == mat_num)).reshape(1))
        mat = props[row]
        stiff_x = mat[1]
        stiff_y = mat[2]
        nu_x = mat[3]
        nu_y = mat[4]
        bulk = mat[5]
        k_l = klocal_transv(
            stiff_x=stiff_x,
            stiff_y=stiff_y,
            nu_x=nu_x,
            nu_y=nu_y,
            bulk=bulk,
            thick=thick,
            length=length,
            b_strip=b_strip,
            b_c=b_c,
            m_i=m_i
        )

        # Transform k_local and kg_local into global coordinates
        alpha = el_props[i, 2]
        gamma = trans(alpha=alpha, total_m=1)
        k_local = gamma @ k_l @ gamma.conj().T

        # Add element contribution of k_local to full matrix k_global and kg_local to kg_global
        node_i = int(elem[1])
        node_j = int(elem[2])
        k_global_transv = assemble_single(
            k_global=k_global_transv,
            k_local=k_local,
            node_i=node_i,
            node_j=node_j,
            n_nodes=n_nodes
        )

    return k_global_transv


def klocal_transv(stiff_x, stiff_y, nu_x, nu_y, bulk, thick, length, b_strip, m_i, b_c):
    #
    # this routine creates the local stiffness matrix for bending terms
    # basically the same way as in the main program, however:
    #   only for single half-wave number m_i
    #   membrane strains practically zero, (membrane moduli are enlarged)
    #   for bending, only transverse terms are considered, (practically: only
    #   keeps the i_1 term, set i_2 through i_5 to be zero)
    # also different from the main program, here only involves one single
    # longitudinal term m_i.
    #
    # input/output data
    #   nodes, elements, props - same as elsewhere throughout this program
    #   k_global_transv - global stiffness matrix (geometric included)
    #
    # Z. Li, Jul 10, 2009

    e_1 = stiff_x / (1 - nu_x*nu_y) * 100000000
    e_2 = stiff_y / (1 - nu_x*nu_y)
    d_x = stiff_x * thick**3 / (12 * (1 - nu_x*nu_y))
    d_y = stiff_y * thick**3 / (12 * (1 - nu_x*nu_y))
    d_1 = nu_x * stiff_y * thick**3 / (12 * (1 - nu_x*nu_y))
    d_xy = bulk * thick**3 / 12

    k_local = np.zeros((8, 8))
    u_m = m_i * np.pi
    u_p = m_i * np.pi
    c_1 = u_m / length
    c_2 = u_p / length

    [i_1, _, _, _, _] = bc_i1_5(b_c=b_c, m_i=m_i, m_j=m_i, length=length)
    i_2 = 0
    i_3 = 0
    i_4 = 0
    i_5 = 0

    k_local[0:4, 0:4] = calc_km_mp(
        e_1=e_1,
        e_2=e_2,
        c_1=c_1,
        c_2=c_2,
        b_strip=b_strip,
        bulk=bulk,
        nu_x=nu_x,
        thick=thick,
        i_1=i_1,
        i_2=i_2,
        i_3=i_3,
        i_4=i_4,
        i_5=i_5
    )
    k_local[4:8, 4:8] = calc_kf_mp(
        d_x=d_x,
        d_y=d_y,
        d_1=d_1,
        d_xy=d_xy,
        b_strip=b_strip,
        i_1=i_1,
        i_2=i_2,
        i_3=i_3,
        i_4=i_4,
        i_5=i_5
    )
    return k_local


def calc_km_mp(e_1, e_2, c_1, c_2, b_strip, bulk, nu_x, thick, i_1, i_2, i_3, i_4, i_5):
    km_mp = np.zeros((4, 4))

    # assemble the matrix of Km_mp (membrane stiffness matrix)
    km_mp[0, 0] = e_1*i_1/b_strip + bulk*b_strip*i_5/3
    km_mp[0, 1] = e_2 * nu_x * (-1 / 2 / c_2) * i_3 - bulk*i_5/2/c_2
    km_mp[0, 2] = -e_1 * i_1 / b_strip + bulk*b_strip*i_5/6
    km_mp[0, 3] = e_2 * nu_x * (-1 / 2 / c_2) * i_3 + bulk*i_5/2/c_2

    km_mp[1, 0] = e_2 * nu_x * (-1 / 2 / c_1) * i_2 - bulk*i_5/2/c_1
    km_mp[1, 1] = e_2*b_strip*i_4/3/c_1/c_2 + bulk*i_5/b_strip/c_1/c_2
    km_mp[1, 2] = e_2 * nu_x * (1/2/c_1) * i_2 - bulk*i_5/2/c_1
    km_mp[1, 3] = e_2*b_strip*i_4/6/c_1/c_2 - bulk*i_5/b_strip/c_1/c_2

    km_mp[2, 0] = -e_1 * i_1 / b_strip + bulk*b_strip*i_5/6
    km_mp[2, 1] = e_2 * nu_x * (1/2/c_2) * i_3 - bulk*i_5/2/c_2
    km_mp[2, 2] = e_1*i_1/b_strip + bulk*b_strip*i_5/3
    km_mp[2, 3] = e_2 * nu_x * (1/2/c_2) * i_3 + bulk*i_5/2/c_2

    km_mp[3, 0] = e_2 * nu_x * (-1 / 2 / c_1) * i_2 + bulk*i_5/2/c_1
    km_mp[3, 1] = e_2*b_strip*i_4/6/c_1/c_2 - bulk*i_5/b_strip/c_1/c_2
    km_mp[3, 2] = e_2 * nu_x * (1/2/c_1) * i_2 + bulk*i_5/2/c_1
    km_mp[3, 3] = e_2*b_strip*i_4/3/c_1/c_2 + bulk*i_5/b_strip/c_1/c_2

    return km_mp * thick


def calc_kf_mp(d_x, d_y, d_1, d_xy, b_strip, i_1, i_2, i_3, i_4, i_5):
    kf_mp = np.zeros((4, 4))

    # assemble the matrix of Kf_mp (flexural stiffness matrix)
    kf_mp[0, 0] = (5040*d_x*i_1 - 504*b_strip**2*d_1*i_2 - 504*b_strip**2*d_1*i_3 \
        + 156*b_strip**4*d_y*i_4 + 2016*b_strip**2*d_xy*i_5)/420/b_strip**3
    kf_mp[0, 1] = (2520*b_strip*d_x*i_1 - 462*b_strip**3*d_1*i_2 - 42*b_strip**3*d_1*i_3 \
        + 22*b_strip**5*d_y*i_4 + 168*b_strip**3*d_xy*i_5)/420/b_strip**3
    kf_mp[0, 2] = (-5040*d_x*i_1 + 504*b_strip**2*d_1*i_2 + 504*b_strip**2*d_1*i_3 \
        + 54*b_strip**4*d_y*i_4 - 2016*b_strip**2*d_xy*i_5)/420/b_strip**3
    kf_mp[0, 3] = (2520*b_strip*d_x*i_1 - 42*b_strip**3*d_1*i_2 - 42*b_strip**3*d_1*i_3 \
        - 13*b_strip**5*d_y*i_4 + 168*b_strip**3*d_xy*i_5)/420/b_strip**3

    kf_mp[1, 0] = (2520*b_strip*d_x*i_1 - 462*b_strip**3*d_1*i_3 - 42*b_strip**3*d_1*i_2 \
        + 22*b_strip**5*d_y*i_4 + 168*b_strip**3*d_xy*i_5)/420/b_strip**3
    kf_mp[1, 1] = (1680*b_strip**2*d_x*i_1 - 56*b_strip**4*d_1*i_2 - 56*b_strip**4*d_1*i_3 \
        + 4*b_strip**6*d_y*i_4 + 224*b_strip**4*d_xy*i_5)/420/b_strip**3
    kf_mp[1, 2] = (-2520*b_strip*d_x*i_1 + 42*b_strip**3*d_1*i_2 + 42*b_strip**3*d_1*i_3 \
        + 13*b_strip**5*d_y*i_4 - 168*b_strip**3*d_xy*i_5)/420/b_strip**3
    kf_mp[1, 3] = (840*b_strip**2*d_x*i_1 + 14*b_strip**4*d_1*i_2 + 14*b_strip**4*d_1*i_3 \
        - 3*b_strip**6*d_y*i_4 - 56*b_strip**4*d_xy*i_5)/420/b_strip**3

    kf_mp[2, 0] = kf_mp[0, 2]
    kf_mp[2, 1] = kf_mp[1, 2]
    kf_mp[2, 2] = (5040*d_x*i_1 - 504*b_strip**2*d_1*i_2 - 504*b_strip**2*d_1*i_3 \
        + 156*b_strip**4*d_y*i_4 + 2016*b_strip**2*d_xy*i_5)/420/b_strip**3
    kf_mp[2, 3] = (-2520*b_strip*d_x*i_1 + 462*b_strip**3*d_1*i_2 + 42*b_strip**3*d_1*i_3 \
        - 22*b_strip**5*d_y*i_4 - 168*b_strip**3*d_xy*i_5)/420/b_strip**3

    kf_mp[3, 0] = kf_mp[0, 3]
    kf_mp[3, 1] = kf_mp[1, 3]
    kf_mp[3, 2] = (-2520*b_strip*d_x*i_1 + 462*b_strip**3*d_1*i_3 + 42*b_strip**3*d_1*i_2 \
        - 22*b_strip**5*d_y*i_4 - 168*b_strip**3*d_xy*i_5)/420/b_strip**3 # not symmetric
    kf_mp[3, 3] = (1680*b_strip**2*d_x*i_1 - 56*b_strip**4*d_1*i_2 - 56*b_strip**4*d_1*i_3 \
        + 4*b_strip**6*d_y*i_4 + 224*b_strip**4*d_xy*i_5)/420/b_strip**3

    return kf_mp


def calc_gm_mp(u_i, u_j, b_strip, length, ty_1, ty_2, i_4, i_5):
    gm_mp = np.zeros((4, 4))

    # assemble the matrix of gm_mp (symmetric membrane stability matrix)
    gm_mp[0, 0] = b_strip * (3*ty_1 + ty_2) * i_5 / 12
    gm_mp[0, 2] = b_strip * (ty_1+ty_2) * i_5 / 12
    gm_mp[2, 0] = gm_mp[0, 2]
    gm_mp[1, 1] = b_strip * length**2 * (3*ty_1 + ty_2) * i_4 / 12 / u_i / u_j
    gm_mp[1, 3] = b_strip * length**2 * (ty_1+ty_2) * i_4 / 12 / u_i / u_j
    gm_mp[3, 1] = gm_mp[1, 3]
    gm_mp[2, 2] = b_strip * (ty_1 + 3*ty_2) * i_5 / 12
    gm_mp[3, 3] = b_strip * length**2 * (ty_1 + 3*ty_2) * i_4 / 12 / u_i / u_j

    return gm_mp


def calc_gf_mp(ty_1, ty_2, b_strip, i_5):
    gf_mp = np.zeros((4, 4))

    # assemble the matrix of gf_mp (symmetric flexural stability matrix)
    gf_mp[0, 0] = (10*ty_1 + 3*ty_2) * b_strip * i_5 / 35
    gf_mp[0, 1] = (15*ty_1 + 7*ty_2) * b_strip**2 * i_5 / 210 / 2
    gf_mp[1, 0] = gf_mp[0, 1]
    gf_mp[0, 2] = 9 * (ty_1+ty_2) * b_strip * i_5 / 140
    gf_mp[2, 0] = gf_mp[0, 2]
    gf_mp[0, 3] = -(7*ty_1 + 6*ty_2) * b_strip**2 * i_5 / 420
    gf_mp[3, 0] = gf_mp[0, 3]
    gf_mp[1, 1] = (5*ty_1 + 3*ty_2) * b_strip**3 * i_5 / 2 / 420
    gf_mp[1, 2] = (6*ty_1 + 7*ty_2) * b_strip**2 * i_5 / 420
    gf_mp[2, 1] = gf_mp[1, 2]
    gf_mp[1, 3] = -(ty_1 + ty_2) * b_strip**3 * i_5 / 140 / 2
    gf_mp[3, 1] = gf_mp[1, 3]
    gf_mp[2, 2] = (3*ty_1 + 10*ty_2) * b_strip * i_5 / 35
    gf_mp[2, 3] = -(7*ty_1 + 15*ty_2) * b_strip**2 * i_5 / 420
    gf_mp[3, 2] = gf_mp[2, 3]
    gf_mp[3, 3] = (3*ty_1 + 5*ty_2) * b_strip**3 * i_5 / 420 / 2

    return gf_mp


def bc_i1_5(b_c, m_i, m_j, length):
    # Calculate the 5 undetermined parameters i_1,i_2,i_3,i_4,i_5 for local elastic
    # and geometric stiffness matrices.
    # b_c: a string specifying boundary conditions to be analysed:
    #'S-S' simply-pimply supported boundary condition at loaded edges
    #'C-C' clamped-clamped boundary condition at loaded edges
    #'S-C' simply-clamped supported boundary condition at loaded edges
    #'C-F' clamped-free supported boundary condition at loaded edges
    #'C-G' clamped-guided supported boundary condition at loaded edges
    # Outputs:
    # i_1,i_2,i_3,i_4,i_5
    # calculation of i_1 is the integration of y_m*Yn from 0 to length
    # calculation of i_2 is the integration of y_m''*Yn from 0 to length
    # calculation of i_3 is the integration of y_m*Yn'' from 0 to length
    # calculation of i_3 is the integration of y_m*Yn'' from 0 to length
    # calculation of i_4 is the integration of y_m''*Yn'' from 0 to length
    # calculation of i_5 is the integration of y_m'*Yn' from 0 to length

    i_1 = 0
    i_2 = 0
    i_3 = 0
    i_4 = 0
    i_5 = 0

    if b_c == 'S-S':
        # For simply-pimply supported boundary condition at loaded edges
        if m_i == m_j:
            i_1 = length / 2
            i_2 = -m_i**2 * np.pi**2 / length / 2
            i_3 = -m_j**2 * np.pi**2 / length / 2
            i_4 = np.pi**4 * m_i**4 / 2 / length**3
            i_5 = np.pi**2 * m_i**2 / 2 / length

    elif b_c == 'C-C':
        # For Clamped-clamped boundary condition at loaded edges
        # calculation of i_1 is the integration of y_m*Yn from 0 to length
        if m_i == m_j:
            if m_i == 1:
                i_1 = 3 * length / 8
            else:
                i_1 = length / 4
            i_2 = -(m_i**2 + 1) * np.pi**2 / 4 / length
            i_3 = -(m_j**2 + 1) * np.pi**2 / 4 / length
            i_4 = np.pi**4 * ((m_i**2 + 1)**2 + 4 * m_i**2) / 4 / length**3
            i_5 = (1 + m_i**2) * np.pi**2 / 4 / length
        else:
            if m_i - m_j == 2:
                i_1 = -length / 8
                i_2 = (m_i**2 + 1) * np.pi**2 / 8 / length - m_i * np.pi**2 / 4 / length
                i_3 = (m_j**2 + 1) * np.pi**2 / 8 / length + m_j * np.pi**2 / 4 / length
                i_4 = -(m_i - 1)**2 * (m_j + 1)**2 * np.pi**4 / 8 / length**3
                i_5 = -(1 + m_i*m_j) * np.pi**2 / 8 / length
            elif m_i - m_j == -2:
                i_1 = -length / 8
                i_2 = (m_i**2 + 1) * np.pi**2 / 8 / length + m_i * np.pi**2 / 4 / length
                i_3 = (m_j**2 + 1) * np.pi**2 / 8 / length - m_j * np.pi**2 / 4 / length
                i_4 = -(m_i + 1)**2 * (m_j - 1)**2 * np.pi**4 / 8 / length**3
                i_5 = -(1 + m_i*m_j) * np.pi**2 / 8 / length

    elif b_c == 'S-C' or b_c == 'C-S':
        # For simply-clamped supported boundary condition at loaded edges
        # calculation of i_1 is the integration of y_m*Yn from 0 to length
        if m_i == m_j:
            i_1 = (1 + (m_i + 1)**2 / m_i**2) * length / 2
            i_2 = -(m_i + 1)**2 * np.pi**2 / length
            i_3 = -(m_i + 1)**2 * np.pi**2 / length
            i_4 = (m_i + 1)**2 * np.pi**4 * ((m_i + 1)**2 + m_i**2) / 2 / length**3
            i_5 = (1 + m_i)**2 * np.pi**2 / length
        else:
            if m_i - m_j == 1:
                i_1 = (m_i+1) * length / 2 / m_i
                i_2 = -(m_i + 1) * m_i * np.pi**2 / 2 / length
                i_3 = -(m_j + 1)**2 * np.pi**2 * (m_i+1) / 2 / length / m_i
                i_4 = (m_i+1) * m_i * (m_j + 1)**2 * np.pi**4 / 2 / length**3
                i_5 = (1+m_i) * (1+m_j) * np.pi**2 / 2 / length
            elif m_i - m_j == -1:
                i_1 = (m_j+1) * length / 2 / m_j
                i_2 = -(m_i + 1)**2 * np.pi**2 * (m_j+1) / 2 / length / m_j
                i_3 = -(m_j + 1) * m_j * np.pi**2 / 2 / length
                i_4 = (m_i + 1)**2 * m_j * (m_j+1) * np.pi**4 / 2 / length**3
                i_5 = (1+m_i) * (1+m_j) * np.pi**2 / 2 / length

    elif b_c == 'C-F' or b_c == 'F-C':
        # For clamped-free supported boundary condition at loaded edges
        # calculation of i_1 is the integration of y_m*Yn from 0 to length
        if m_i == m_j:
            i_1 = 3*length/2 - 2 * length * (-1)**(m_i - 1) / (m_i - 1/2) / np.pi
            i_2 = (m_i - 1/2)**2 * np.pi**2 * ((-1)**(m_i - 1) / (m_i - 1/2) / np.pi - 1/2) / length
            i_3 = (m_j - 1/2)**2 * np.pi**2 * ((-1)**(m_j - 1) / (m_j - 1/2) / np.pi - 1/2) / length
            i_4 = (m_i - 1/2)**4 * np.pi**4 / 2 / length**3
            i_5 = (m_i - 1/2)**2 * np.pi**2 / 2 / length
        else:
            i_1 = length - length*(-1)**(m_i - 1) / (m_i - 1/2) / np.pi \
                - length*(-1)**(m_j - 1) / (m_j - 1/2) / np.pi
            i_2 = (m_i - 1/2)**2 * np.pi**2 * ((-1)**(m_i - 1) / (m_i - 1/2) / np.pi) / length
            i_3 = (m_j - 1/2)**2 * np.pi**2 * ((-1)**(m_j - 1) / (m_j - 1/2) / np.pi) / length

    elif b_c == 'C-G' or b_c == 'G-C':
        # For clamped-guided supported boundary condition at loaded edges
        # calculation of i_1 is the integration of y_m*Yn from 0 to length
        if m_i == m_j:
            if m_i == 1:
                i_1 = 3 * length / 8
            else:
                i_1 = length / 4
            i_2 = -((m_i - 1/2)**2 + 1/4) * np.pi**2 / length / 4
            i_3 = -((m_i - 1/2)**2 + 1/4) * np.pi**2 / length / 4
            i_4 = ((m_i - 1/2)**2
                   + 1/4)**2 * np.pi**4 / 4 / length**3 + (m_i - 1/2)**2 * np.pi**4 / 4 / length**3
            i_5 = (m_i - 1/2)**2 * np.pi**2 / length / 4 + np.pi**2 / 16 / length
        else:
            if m_i - m_j == 1:
                i_1 = -length / 8
                i_2 = ((m_i - 1/2)**2
                       + 1/4) * np.pi**2 / length / 8 - (m_i - 1/2) * np.pi**2 / length / 8
                i_3 = ((m_j - 1/2)**2
                       + 1/4) * np.pi**2 / length / 8 + (m_j - 1/2) * np.pi**2 / length / 8
                i_4 = -m_j**4 * np.pi**4 / 8 / length**3
                i_5 = -m_j**2 * np.pi**2 / 8 / length
            elif m_i - m_j == -1:
                i_1 = -length / 8
                i_2 = ((m_i - 1/2)**2
                       + 1/4) * np.pi**2 / length / 8 + (m_i - 1/2) * np.pi**2 / length / 8
                i_3 = ((m_j - 1/2)**2
                       + 1/4) * np.pi**2 / length / 8 - (m_j - 1/2) * np.pi**2 / length / 8
                i_4 = -m_i**4 * np.pi**4 / 8 / length**3
                i_5 = -m_i**2 * np.pi**2 / 8 / length

    return [i_1, i_2, i_3, i_4, i_5]


def trans(alpha, total_m):
    # Transfer the local stiffness into global stiffness
    # Zhanjie 2008
    # modified by Z. Li, Aug. 09, 2009

    gamma = np.zeros((8 * total_m, 8 * total_m))

    gam = np.array([[np.cos(alpha), 0, 0, 0, -np.sin(alpha), 0, 0, 0], [0, 1, 0, 0, 0, 0, 0, 0],
                    [0, 0, np.cos(alpha), 0, 0, 0, -np.sin(alpha), 0], [0, 0, 0, 1, 0, 0, 0, 0],
                    [np.sin(alpha), 0, 0, 0, np.cos(alpha), 0, 0, 0], [0, 0, 0, 0, 0, 1, 0, 0],
                    [0, 0, np.sin(alpha), 0, 0, 0, np.cos(alpha), 0], [0, 0, 0, 0, 0, 0, 0, 1]])

    # extend to multi-m
    for i in range(0, total_m):
        gamma[8 * i:8 * (i+1), 8 * i:8 * (i+1)] = gam

    return gamma


def assemble(k_global, kg_global, k_local, kg_local, node_i, node_j, n_nodes, m_a):
    # Add the element contribution to the global stiffness matrix

    # Outputs:
    # k_global: global elastic stiffness matrix
    # kg_global: global geometric stiffness matrix
    # k_global and kg_global: total_m x total_m submatrices. Each submatrix is similar to the
    # one used in original CUFSM for single longitudinal term m in the DOF order
    #[u1 v1...un vn w1 01...wn 0n]m'.

    # Z. Li, June 2008
    # modified by Z. Li, Aug. 09, 2009
    # Z. Li, June 2010

    total_m = len(m_a)  # Total number of longitudinal terms m
    skip = 2 * n_nodes
    for i in range(0, total_m):
        for j in range(0, total_m):
            # Submatrices for the initial stiffness
            k11 = k_local[8 * i:8*i + 2, 8 * j:8*j + 2]
            k12 = k_local[8 * i:8*i + 2, 8*j + 2:8*j + 4]
            k13 = k_local[8 * i:8*i + 2, 8*j + 4:8*j + 6]
            k14 = k_local[8 * i:8*i + 2, 8*j + 6:8*j + 8]
            k21 = k_local[8*i + 2:8*i + 4, 8 * j:8*j + 2]
            k22 = k_local[8*i + 2:8*i + 4, 8*j + 2:8*j + 4]
            k23 = k_local[8*i + 2:8*i + 4, 8*j + 4:8*j + 6]
            k24 = k_local[8*i + 2:8*i + 4, 8*j + 6:8*j + 8]
            k31 = k_local[8*i + 4:8*i + 6, 8 * j:8*j + 2]
            k32 = k_local[8*i + 4:8*i + 6, 8*j + 2:8*j + 4]
            k33 = k_local[8*i + 4:8*i + 6, 8*j + 4:8*j + 6]
            k34 = k_local[8*i + 4:8*i + 6, 8*j + 6:8*j + 8]
            k41 = k_local[8*i + 6:8*i + 8, 8 * j:8*j + 2]
            k42 = k_local[8*i + 6:8*i + 8, 8*j + 2:8*j + 4]
            k43 = k_local[8*i + 6:8*i + 8, 8*j + 4:8*j + 6]
            k44 = k_local[8*i + 6:8*i + 8, 8*j + 6:8*j + 8]

            k_global[4*n_nodes*i+(node_i+1)*2-2:4*n_nodes*i+(node_i+1)*2, \
                4*n_nodes*j+(node_i+1)*2-2:4*n_nodes*j+(node_i+1)*2] += k11
            k_global[4*n_nodes*i+(node_i+1)*2-2:4*n_nodes*i+(node_i+1)*2, \
                4*n_nodes*j+(node_j+1)*2-2:4*n_nodes*j+(node_j+1)*2] += k12
            k_global[4*n_nodes*i+(node_j+1)*2-2:4*n_nodes*i+(node_j+1)*2, \
                4*n_nodes*j+(node_i+1)*2-2:4*n_nodes*j+(node_i+1)*2] += k21
            k_global[4*n_nodes*i+(node_j+1)*2-2:4*n_nodes*i+(node_j+1)*2, \
                4*n_nodes*j+(node_j+1)*2-2:4*n_nodes*j+(node_j+1)*2] += k22

            k_global[4*n_nodes*i+skip+(node_i+1)*2-2:4*n_nodes*i+skip+(node_i+1)*2, \
                4*n_nodes*j+skip+(node_i+1)*2-2:4*n_nodes*j+skip+(node_i+1)*2] += k33
            k_global[4*n_nodes*i+skip+(node_i+1)*2-2:4*n_nodes*i+skip+(node_i+1)*2, \
                4*n_nodes*j+skip+(node_j+1)*2-2:4*n_nodes*j+skip+(node_j+1)*2] += k34
            k_global[4*n_nodes*i+skip+(node_j+1)*2-2:4*n_nodes*i+skip+(node_j+1)*2, \
                4*n_nodes*j+skip+(node_i+1)*2-2:4*n_nodes*j+skip+(node_i+1)*2] += k43
            k_global[4*n_nodes*i+skip+(node_j+1)*2-2:4*n_nodes*i+skip+(node_j+1)*2, \
                4*n_nodes*j+skip+(node_j+1)*2-2:4*n_nodes*j+skip+(node_j+1)*2] += k44

            k_global[4*n_nodes*i+(node_i+1)*2-2:4*n_nodes*i+(node_i+1)*2, \
                4*n_nodes*j+skip+(node_i+1)*2-2:4*n_nodes*j+skip+(node_i+1)*2] += k13
            k_global[4*n_nodes*i+(node_i+1)*2-2:4*n_nodes*i+(node_i+1)*2, \
                4*n_nodes*j+skip+(node_j+1)*2-2:4*n_nodes*j+skip+(node_j+1)*2] += k14
            k_global[4*n_nodes*i+(node_j+1)*2-2:4*n_nodes*i+(node_j+1)*2, \
                4*n_nodes*j+skip+(node_i+1)*2-2:4*n_nodes*j+skip+(node_i+1)*2] += k23
            k_global[4*n_nodes*i+(node_j+1)*2-2:4*n_nodes*i+(node_j+1)*2, \
                4*n_nodes*j+skip+(node_j+1)*2-2:4*n_nodes*j+skip+(node_j+1)*2] += k24

            k_global[4*n_nodes*i+skip+(node_i+1)*2-2:4*n_nodes*i+skip+(node_i+1)*2, \
                4*n_nodes*j+(node_i+1)*2-2:4*n_nodes*j+(node_i+1)*2] += k31
            k_global[4*n_nodes*i+skip+(node_i+1)*2-2:4*n_nodes*i+skip+(node_i+1)*2, \
                4*n_nodes*j+(node_j+1)*2-2:4*n_nodes*j+(node_j+1)*2] += k32
            k_global[4*n_nodes*i+skip+(node_j+1)*2-2:4*n_nodes*i+skip+(node_j+1)*2, \
                4*n_nodes*j+(node_i+1)*2-2:4*n_nodes*j+(node_i+1)*2] += k41
            k_global[4*n_nodes*i+skip+(node_j+1)*2-2:4*n_nodes*i+skip+(node_j+1)*2, \
                4*n_nodes*j+(node_j+1)*2-2:4*n_nodes*j+(node_j+1)*2] += k42

            # Submatrices for the initial stiffness
            kg11 = kg_local[8 * i:8*i + 2, 8 * j:8*j + 2]
            kg12 = kg_local[8 * i:8*i + 2, 8*j + 2:8*j + 4]
            kg13 = kg_local[8 * i:8*i + 2, 8*j + 4:8*j + 6]
            kg14 = kg_local[8 * i:8*i + 2, 8*j + 6:8*j + 8]
            kg21 = kg_local[8*i + 2:8*i + 4, 8 * j:8*j + 2]
            kg22 = kg_local[8*i + 2:8*i + 4, 8*j + 2:8*j + 4]
            kg23 = kg_local[8*i + 2:8*i + 4, 8*j + 4:8*j + 6]
            kg24 = kg_local[8*i + 2:8*i + 4, 8*j + 6:8*j + 8]
            kg31 = kg_local[8*i + 4:8*i + 6, 8 * j:8*j + 2]
            kg32 = kg_local[8*i + 4:8*i + 6, 8*j + 2:8*j + 4]
            kg33 = kg_local[8*i + 4:8*i + 6, 8*j + 4:8*j + 6]
            kg34 = kg_local[8*i + 4:8*i + 6, 8*j + 6:8*j + 8]
            kg41 = kg_local[8*i + 6:8*i + 8, 8 * j:8*j + 2]
            kg42 = kg_local[8*i + 6:8*i + 8, 8*j + 2:8*j + 4]
            kg43 = kg_local[8*i + 6:8*i + 8, 8*j + 4:8*j + 6]
            kg44 = kg_local[8*i + 6:8*i + 8, 8*j + 6:8*j + 8]

            kg_global[4*n_nodes*i + (node_i+1) * 2 - 2:4*n_nodes*i + (node_i+1) * 2,
                      4*n_nodes*j + (node_i+1) * 2 - 2:4*n_nodes*j + (node_i+1) * 2] += kg11
            kg_global[4*n_nodes*i + (node_i+1) * 2 - 2:4*n_nodes*i + (node_i+1) * 2,
                      4*n_nodes*j + (node_j+1) * 2 - 2:4*n_nodes*j + (node_j+1) * 2] += kg12
            kg_global[4*n_nodes*i + (node_j+1) * 2 - 2:4*n_nodes*i + (node_j+1) * 2,
                      4*n_nodes*j + (node_i+1) * 2 - 2:4*n_nodes*j + (node_i+1) * 2] += kg21
            kg_global[4*n_nodes*i + (node_j+1) * 2 - 2:4*n_nodes*i + (node_j+1) * 2,
                      4*n_nodes*j + (node_j+1) * 2 - 2:4*n_nodes*j + (node_j+1) * 2] += kg22

            kg_global[4*n_nodes*i + skip + (node_i+1) * 2 - 2:4*n_nodes*i + skip + (node_i+1) * 2,
                      4*n_nodes*j + skip + (node_i+1) * 2 - 2:4*n_nodes*j + skip
                      + (node_i+1) * 2] += kg33
            kg_global[4*n_nodes*i + skip + (node_i+1) * 2 - 2:4*n_nodes*i + skip + (node_i+1) * 2,
                      4*n_nodes*j + skip + (node_j+1) * 2 - 2:4*n_nodes*j + skip
                      + (node_j+1) * 2] += kg34
            kg_global[4*n_nodes*i + skip + (node_j+1) * 2 - 2:4*n_nodes*i + skip + (node_j+1) * 2,
                      4*n_nodes*j + skip + (node_i+1) * 2 - 2:4*n_nodes*j + skip
                      + (node_i+1) * 2] += kg43
            kg_global[4*n_nodes*i + skip + (node_j+1) * 2 - 2:4*n_nodes*i + skip + (node_j+1) * 2,
                      4*n_nodes*j + skip + (node_j+1) * 2 - 2:4*n_nodes*j + skip
                      + (node_j+1) * 2] += kg44

            kg_global[4*n_nodes*i + (node_i+1) * 2 - 2:4*n_nodes*i + (node_i+1) * 2, 4*n_nodes*j
                      + skip + (node_i+1) * 2 - 2:4*n_nodes*j + skip + (node_i+1) * 2] += kg13
            kg_global[4*n_nodes*i + (node_i+1) * 2 - 2:4*n_nodes*i + (node_i+1) * 2, 4*n_nodes*j
                      + skip + (node_j+1) * 2 - 2:4*n_nodes*j + skip + (node_j+1) * 2] += kg14
            kg_global[4*n_nodes*i + (node_j+1) * 2 - 2:4*n_nodes*i + (node_j+1) * 2, 4*n_nodes*j
                      + skip + (node_i+1) * 2 - 2:4*n_nodes*j + skip + (node_i+1) * 2] += kg23
            kg_global[4*n_nodes*i + (node_j+1) * 2 - 2:4*n_nodes*i + (node_j+1) * 2, 4*n_nodes*j
                      + skip + (node_j+1) * 2 - 2:4*n_nodes*j + skip + (node_j+1) * 2] += kg24

            kg_global[4*n_nodes*i + skip + (node_i+1) * 2 - 2:4*n_nodes*i + skip + (node_i+1) * 2,
                      4*n_nodes*j + (node_i+1) * 2 - 2:4*n_nodes*j + (node_i+1) * 2] += kg31
            kg_global[4*n_nodes*i + skip + (node_i+1) * 2 - 2:4*n_nodes*i + skip + (node_i+1) * 2,
                      4*n_nodes*j + (node_j+1) * 2 - 2:4*n_nodes*j + (node_j+1) * 2] += kg32
            kg_global[4*n_nodes*i + skip + (node_j+1) * 2 - 2:4*n_nodes*i + skip + (node_j+1) * 2,
                      4*n_nodes*j + (node_i+1) * 2 - 2:4*n_nodes*j + (node_i+1) * 2] += kg41
            kg_global[4*n_nodes*i + skip + (node_j+1) * 2 - 2:4*n_nodes*i + skip + (node_j+1) * 2,
                      4*n_nodes*j + (node_j+1) * 2 - 2:4*n_nodes*j + (node_j+1) * 2] += kg42

    return k_global, kg_global


def assemble_single(k_global, k_local, node_i, node_j, n_nodes):
    #
    # this routine adds the element contribution to the global stiffness matrix
    # basically it does the same as routine 'assemble', however:
    #   it does not care about kg_global (geom stiff matrix)
    #   only involves single half-wave number m_i

    # S. Adany, Feb 06, 2004
    # Z. Li, Jul 10, 2009
    #
    # submatrices for the initial stiffness
    k11 = k_local[0:2, 0:2]
    k12 = k_local[0:2, 2:4]
    k13 = k_local[0:2, 4:6]
    k14 = k_local[0:2, 6:8]
    k21 = k_local[2:4, 0:2]
    k22 = k_local[2:4, 2:4]
    k23 = k_local[2:4, 4:6]
    k24 = k_local[2:4, 6:8]
    k31 = k_local[4:6, 0:2]
    k32 = k_local[4:6, 2:4]
    k33 = k_local[4:6, 4:6]
    k34 = k_local[4:6, 6:8]
    k41 = k_local[6:8, 0:2]
    k42 = k_local[6:8, 2:4]
    k43 = k_local[6:8, 4:6]
    k44 = k_local[6:8, 6:8]

    # the additional terms for k_global are stored in k_2_matrix
    skip = 2 * n_nodes
    k_global[node_i * 2:node_i*2 + 2, node_i * 2:node_i*2 + 2] += k11
    k_global[node_i * 2:node_i*2 + 2, node_j * 2:node_j*2 + 2] += k12
    k_global[node_j * 2:node_j*2 + 2, node_i * 2:node_i*2 + 2] += k21
    k_global[node_j * 2:node_j*2 + 2, node_j * 2:node_j*2 + 2] += k22

    k_global[skip + node_i*2:skip + node_i*2 + 2, skip + node_i*2:skip + node_i*2 + 2] += k33
    k_global[skip + node_i*2:skip + node_i*2 + 2, skip + node_j*2:skip + node_j*2 + 2] += k34
    k_global[skip + node_j*2:skip + node_j*2 + 2, skip + node_i*2:skip + node_i*2 + 2] += k43
    k_global[skip + node_j*2:skip + node_j*2 + 2, skip + node_j*2:skip + node_j*2 + 2] += k44

    k_global[node_i * 2:node_i*2 + 2, skip + node_i*2:skip + node_i*2 + 2] += k13
    k_global[node_i * 2:node_i*2 + 2, skip + node_j*2:skip + node_j*2 + 2] += k14
    k_global[node_j * 2:node_j*2 + 2, skip + node_i*2:skip + node_i*2 + 2] += k23
    k_global[node_j * 2:node_j*2 + 2, skip + node_j*2:skip + node_j*2 + 2] += k24

    k_global[skip + node_i*2:skip + node_i*2 + 2, node_i * 2:node_i*2 + 2] += k31
    k_global[skip + node_i*2:skip + node_i*2 + 2, node_j * 2:node_j*2 + 2] += k32
    k_global[skip + node_j*2:skip + node_j*2 + 2, node_i * 2:node_i*2 + 2] += k41
    k_global[skip + node_j*2:skip + node_j*2 + 2, node_j * 2:node_j*2 + 2] += k42

    return k_global


def spring_klocal(k_u, k_v, k_w, k_q, length, b_c, m_a, discrete, y_s):
    # Generate spring stiffness matrix (k_local) in local coordinates, modified from
    # klocal
    # BWS DEC 2015

    # Inputs:
    # k_u,k_v,k_w,k_q spring stiffness values
    # length: length of the strip in longitudinal direction
    # b_c: ['S-S'] a string specifying boundary conditions to be analyzed:
    #'S-S' simply-pimply supported boundary condition at loaded edges
    #'C-C' clamped-clamped boundary condition at loaded edges
    #'S-C' simply-clamped supported boundary condition at loaded edges
    #'C-F' clamped-free supported boundary condition at loaded edges
    #'C-G' clamped-guided supported boundary condition at loaded edges
    # m_a: longitudinal terms (or half-wave numbers) for this length
    # discrete == 1 if discrete spring
    # y_s = location of discrete spring

    # Output:
    # k_local: local stiffness matrix, a total_m x total_m matrix of 8 by 8 submatrices.
    # k_local=[k_mp]total_m x total_m block matrix
    # each k_mp is the 8 x 8 submatrix in the DOF order [u1 v1 u2 v2 w1 theta1 w2 theta2]'

    total_m = len(m_a)  # Total number of longitudinal terms m
    k_local = np.zeros(8 * total_m, 8 * total_m)
    for i in range(0, total_m):
        for j in range(0, total_m):
            km_mp = np.zeros(4, 4)
            kf_mp = np.zeros(4, 4)
            u_i = m_a[i] * np.pi
            u_j = m_a[j] * np.pi

            if discrete:
                [i_1, i_5] = bc_i1_5_atpoint(
                    b_c=b_c, m_i=m_a[i], m_j=m_a[j], length=length, y_s=y_s
                )
            else:  # foundation spring
                [i_1, _, _, _, i_5] = bc_i1_5(b_c=b_c, m_i=m_a[i], m_j=m_a[j], length=length)
            # assemble the matrix of km_mp (membrane stiffness)
            km_mp = np.array(
                [[k_u * i_1, 0, -k_u * i_1, 0],
                 [0, k_v * i_5 * length**2 / (u_i*u_j), 0, -k_v * i_5 * length**2 / (u_i*u_j)],
                 [-k_u * i_1, 0, k_u * i_1, 0],
                 [0, -k_v * i_5 * length**2 / (u_i*u_j), 0, k_v * i_5 * length**2 / (u_i*u_j)]]
            )
            # assemble the matrix of kf_mp (flexural stiffness)
            kf_mp = np.array([[k_w * i_1, 0, -k_w * i_1, 0], [0, k_q * i_1, 0, -k_q * i_1],
                              [-k_w * i_1, 0, k_w * i_1, 0], [0, -k_q * i_1, 0, k_q * i_1]])

            k_local[8 * i:8*i + 4, 8 * j:8*j + 4] = km_mp
            k_local[8*i + 4:8 * (i+1), 8*j + 4:8 * (j+1)] = kf_mp

    return k_local


def bc_i1_5_atpoint(b_c, m_i, m_j, length, y_s):
    # Calculate the value of the longitudinal shape functions for discrete springs

    # b_c: a string specifying boundary conditions to be analyzed:
    #'S-S' simply-pimply supported boundary condition at loaded edges
    #'C-C' clamped-clamped boundary condition at loaded edges
    #'S-C' simply-clamped supported boundary condition at loaded edges
    #'C-F' clamped-free supported boundary condition at loaded edges
    #'C-G' clamped-guided supported boundary condition at loaded edges
    # Outputs:
    # i_1,i_5
    # calculation of i_1 is the value of y_m(y/L)*Yn(y/L)
    # calculation of i_5 is the value of y_m'(y/L)*Yn'(y/L)

    y_i = ym_at_ys(b_c=b_c, m_i=m_i, y_s=y_s, length=length)
    y_j = ym_at_ys(b_c=b_c, m_i=m_j, y_s=y_s, length=length)
    y_i_prime = ymprime_at_ys(b_c=b_c, m_i=m_i, y_s=y_s, length=length)
    y_j_prime = ymprime_at_ys(b_c=b_c, m_i=m_j, y_s=y_s, length=length)
    i_1 = y_i * y_j
    i_5 = y_i_prime * y_j_prime
    return i_1, i_5


def spring_assemble(k_global, k_local, node_i, node_j, n_nodes, m_a):
    # Add the (spring) contribution to the global stiffness matrix

    # Outputs:
    # k_global: global elastic stiffness matrix
    # k_global and Kg: total_m x total_m submatrices. Each submatrix is similar to the
    # one used in original CUFSM for single longitudinal term m in the DOF order
    #[u1 v1...un vn w1 01...wn 0n]m'.

    # Z. Li, June 2008
    # modified by Z. Li, Aug. 09, 2009
    # Z. Li, June 2010
    # adapted for springs BWS Dec 2015

    total_m = len(m_a)  # Total number of longitudinal terms m
    skip = 2 * n_nodes
    for i in range(0, total_m):
        for j in range(0, total_m):
            # Submatrices for the initial stiffness
            k11 = k_local[8 * i:8*i + 2, 8 * j:8*j + 2]
            k12 = k_local[8 * i:8*i + 2, 8*j + 2:8*j + 4]
            k13 = k_local[8 * i:8*i + 2, 8*j + 4:8*j + 6]
            k14 = k_local[8 * i:8*i + 2, 8*j + 6:8*j + 8]
            k21 = k_local[8*i + 2:8*i + 4, 8 * j:8*j + 2]
            k22 = k_local[8*i + 2:8*i + 4, 8*j + 2:8*j + 4]
            k23 = k_local[8*i + 2:8*i + 4, 8*j + 4:8*j + 6]
            k24 = k_local[8*i + 2:8*i + 4, 8*j + 6:8*j + 8]
            k31 = k_local[8*i + 4:8*i + 6, 8 * j:8*j + 2]
            k32 = k_local[8*i + 4:8*i + 6, 8*j + 2:8*j + 4]
            k33 = k_local[8*i + 4:8*i + 6, 8*j + 4:8*j + 6]
            k34 = k_local[8*i + 4:8*i + 6, 8*j + 6:8*j + 8]
            k41 = k_local[8*i + 6:8*i + 8, 8 * j:8*j + 2]
            k42 = k_local[8*i + 6:8*i + 8, 8*j + 2:8*j + 4]
            k43 = k_local[8*i + 6:8*i + 8, 8*j + 4:8*j + 6]
            k44 = k_local[8*i + 6:8*i + 8, 8*j + 6:8*j + 8]

            k_global[4*n_nodes*i + (node_i+1) * 2 - 1:4*n_nodes*i + (node_i+1) * 2,
                     4*n_nodes*j + (node_i+1) * 2 - 1:4*n_nodes*j + (node_i+1) * 2] += k11
            if node_j != 0:
                k_global[4*n_nodes*i + (node_i+1) * 2 - 1:4*n_nodes*i + (node_i+1) * 2,
                         4*n_nodes*j + (node_j+1) * 2 - 1:4*n_nodes*j + (node_j+1) * 2] += k12
                k_global[4*n_nodes*i + (node_j+1) * 2 - 1:4*n_nodes*i + (node_j+1) * 2,
                         4*n_nodes*j + (node_i+1) * 2 - 1:4*n_nodes*j + (node_i+1) * 2] += k21
                k_global[4*n_nodes*i + (node_j+1) * 2 - 1:4*n_nodes*i + (node_j+1) * 2,
                         4*n_nodes*j + (node_j+1) * 2 - 1:4*n_nodes*j + (node_j+1) * 2] += k22

            k_global[4*n_nodes*i + skip + (node_i+1) * 2 - 1:4*n_nodes*i + skip + (node_i+1) * 2,
                     4*n_nodes*j + skip + (node_i+1) * 2 - 1:4*n_nodes*j + skip
                     + (node_i+1) * 2] += k33
            if node_j != 0:
                k_global[4*n_nodes*i + skip + (node_i+1) * 2 - 1:4*n_nodes*i + skip
                         + (node_i+1) * 2, 4*n_nodes*j + skip + (node_j+1) * 2 - 1:4*n_nodes*j
                         + skip + (node_j+1) * 2] += k34
                k_global[4*n_nodes*i + skip + (node_j+1) * 2 - 1:4*n_nodes*i + skip
                         + (node_j+1) * 2, 4*n_nodes*j + skip + (node_i+1) * 2 - 1:4*n_nodes*j
                         + skip + (node_i+1) * 2] += k43
                k_global[4*n_nodes*i + skip + (node_j+1) * 2 - 1:4*n_nodes*i + skip
                         + (node_j+1) * 2, 4*n_nodes*j + skip + (node_j+1) * 2 - 1:4*n_nodes*j
                         + skip + (node_j+1) * 2] += k44

            k_global[4*n_nodes*i + (node_i+1) * 2 - 1:4*n_nodes*i + (node_i+1) * 2, 4*n_nodes*j
                     + skip + (node_i+1) * 2 - 1:4*n_nodes*j + skip + (node_i+1) * 2] += k13
            if node_j != 0:
                k_global[4*n_nodes*i + (node_i+1) * 2 - 1:4*n_nodes*i + (node_i+1) * 2, 4*n_nodes*j
                         + skip + (node_j+1) * 2 - 1:4*n_nodes*j + skip + (node_j+1) * 2] += k14
                k_global[4*n_nodes*i + (node_j+1) * 2 - 1:4*n_nodes*i + (node_j+1) * 2, 4*n_nodes*j
                         + skip + (node_i+1) * 2 - 1:4*n_nodes*j + skip + (node_i+1) * 2] += k23
                k_global[4*n_nodes*i + (node_j+1) * 2 - 1:4*n_nodes*i + (node_j+1) * 2, 4*n_nodes*j
                         + skip + (node_j+1) * 2 - 1:4*n_nodes*j + skip + (node_j+1) * 2] += k24

            k_global[4*n_nodes*i + skip + (node_i+1) * 2 - 1:4*n_nodes*i + skip + (node_i+1) * 2,
                     4*n_nodes*j + (node_i+1) * 2 - 1:4*n_nodes*j + (node_i+1) * 2] += k31
            if node_j != 0:
                k_global[4*n_nodes*i + skip + (node_i+1) * 2 - 1:4*n_nodes*i + skip
                         + (node_i+1) * 2,
                         4*n_nodes*j + (node_j+1) * 2 - 1:4*n_nodes*j + (node_j+1) * 2] += k32
                k_global[4*n_nodes*i + skip + (node_j+1) * 2 - 1:4*n_nodes*i + skip
                         + (node_j+1) * 2,
                         4*n_nodes*j + (node_i+1) * 2 - 1:4*n_nodes*j + (node_i+1) * 2] += k41
                k_global[4*n_nodes*i + skip + (node_j+1) * 2 - 1:4*n_nodes*i + skip
                         + (node_j+1) * 2,
                         4*n_nodes*j + (node_j+1) * 2 - 1:4*n_nodes*j + (node_j+1) * 2] += k42

    return k_global


def ym_at_ys(b_c, m_i, y_s, length):
    # Longitudinal shape function values
    # written by BWS in 2015
    # could be called in lots of places,  but now (2015) is hardcoded by Zhanjie
    # in several places in the interface
    # written in 2015 because wanted it for a new idea on discrete springs

    if b_c == 'S-S':
        y_m = np.sin(m_i * np.pi * y_s / length)
    elif b_c == 'C-C':
        y_m = np.sin(m_i * np.pi * y_s / length) * np.sin(np.pi * y_s / length)
    elif b_c == 'S-C' or b_c == 'C-S':
        y_m = np.sin((m_i+1) * np.pi * y_s / length
                     ) + (m_i+1) / m_i * np.sin(m_i * np.pi * y_s / length)
    elif b_c == 'C-F' or b_c == 'F-C':
        y_m = 1 - np.cos((m_i-0.5) * np.pi * y_s / length)
    elif b_c == 'C-G' or b_c == 'G-C':
        y_m = np.sin((m_i-0.5) * np.pi * y_s / length) * np.sin(np.pi * y_s / length / 2)

    return y_m


def ymprime_at_ys(b_c, m_i, y_s, length):
    # First Derivative of Longitudinal shape function values
    # written by BWS in 2015
    # could be called in lots of places,  but now (2015) is hardcoded by Zhanjie
    # in several places in the interface
    # written in 2015 because wanted it for a new idea on discrete springs

    if b_c == 'S-S':
        y_m_prime = (np.pi * m_i * np.cos((np.pi * m_i * y_s) / length)) / length
    elif b_c == 'C-C':
        y_m_prime = (np.pi * np.cos((np.pi*y_s) / length) \
            * np.sin((np.pi*m_i*y_s) / length)) / length \
            + (np.pi*m_i * np.sin((np.pi*y_s)/length) \
                * np.cos((np.pi*m_i*y_s)/length)) / length
    elif b_c == 'S-C' or b_c == 'C-S':
        y_m_prime = (np.pi * np.cos((np.pi*y_s * (m_i + 1))/length) * (m_i + 1)) / length \
            + (np.pi * np.cos((np.pi*m_i*y_s)/length)*(m_i + 1)) / length
    elif b_c == 'C-F' or b_c == 'F-C':
        y_m_prime = (np.pi * np.sin((np.pi * y_s * (m_i - 1/2)) / length) * (m_i - 1/2)) / length
    elif b_c == 'C-G' or b_c == 'G-C':
        y_m_prime = (np.pi*np.sin((np.pi*y_s * (m_i - 1/2))/length) \
            * np.cos((np.pi*y_s)/(2*length)))/(2*length) \
            + (np.pi*np.cos((np.pi*y_s*(m_i - 1/2))/length) \
            * np.sin((np.pi*y_s)/(2*length))*(m_i - 1/2))/length

    return y_m_prime



from copy import deepcopy
from scipy import linalg as spla
import numpy as np
from pycufsm.analysis import analysis
import pycufsm.cfsm
# from scipy.sparse.linalg import eigs
# Originally developed for MATLAB by Benjamin Schafer PhD et al
# Ported to Python by Brooks Smith MEng, PE, CPEng
#
# Each function within this file was originally its own separate file.
# Original MATLAB comments, especially those retaining to authorship or
# change history, have been generally retained unaltered


def strip(
    props, nodes, elements, lengths, springs, constraints, gbt_con, b_c, m_all, n_eigs, sect_props
):

    # INPUTS
    # props: [mat_num stiff_x stiff_y nu_x nu_y bulk] 6 x n_mats
    # nodes: [node# x y dof_x dof_y dof_z dof_r stress] n_nodes x 8
    # elements: [elem# node_i node_j thick mat_num] n_elements x 5
    # lengths: [L1 L2 L3...] 1 x n_lengths lengths to be analysed
    # could be half-wavelengths for signature curve
    # or physical lengths for general b.c.
    # springs: [node# d.o.f. k_spring k_flag] where 1=x dir 2= y dir 3 = z dir 4 = q dir (twist)
    #     flag says if k_stiff is a foundation stiffness or a total stiffness
    # constraints:: [node# e dof_e coeff node# k dof_k] e=dof to be eliminated
    #     k=kept dof dof_e_node = coeff*dof_k_node_k
    # gbt_con: gbt_con.glob,gbt_con.dist, gbt_con.local, gbt_con.other vectors of 1's
    #  and 0's referring to the inclusion (1) or exclusion of a given mode from the analysis,
    #  gbt_con.o_space - choices of ST/O mode
    #         1: ST basis
    #         2: O space (null space of GDL) with respect to k_global
    #         3: O space (null space of GDL) with respect to kg_global
    #         4: O space (null space of GDL) in vector sense
    #  gbt_con.norm - code for normalization (if normalization is done at all)
    #         0: no normalization,
    #         1: vector norm
    #         2: strain energy norm
    #         3: work norm
    #  gbt_con.couple - coupled basis vs uncoupled basis for general
    #             B.C. especially for non-simply supported B.C.
    #         1: uncoupled basis, the basis will be block diagonal
    #         2: coupled basis, the basis is fully spanned
    #  gbt_con.orth - natural basis vs modal basis
    #         1: natural basis
    #         2: modal basis, axial orthogonality
    #         3: modal basis, load dependent orthogonality
    # b_c: ['S-S'] a string specifying boundary conditions to be analysed:
    # 'S-S' simply-pimply supported boundary condition at loaded edges
    # 'C-C' clamped-clamped boundary condition at loaded edges
    # 'S-C' simply-clamped supported boundary condition at loaded edges
    # 'C-F' clamped-free supported boundary condition at loaded edges
    # 'C-G' clamped-guided supported boundary condition at loaded edges
    # m_all: m_all{length#}=[longitudinal_num# ... longitudinal_num#],
    #       longitudinal terms m for all the lengths in cell notation
    # each cell has a vector including the longitudinal terms for this length
    # n_eigs - the number of eigenvalues to be determined at length (default=10)

    # OUTPUTS
    # curve: buckling curve (load factor) for each length
    # curve{i} = [ length mode# 1
    #             length mode# 2
    #             ...    ...
    #             length mode#]
    # shapes = mode shapes for each length
    # shapes{i} = mode, mode is a matrix, each column corresponds to a mode.

    n_nodes = len(nodes)
    curve = []
    shapes = []
    signature = np.zeros((len(lengths), 1))

    # CLEAN UP INPUT
    # clean u_j 0's, multiple terms. or out-of-order terms in m_all
    m_all = analysis.m_sort(m_all)

    # DETERMINE FLAGS FOR USER CONSTRAINTS AND INTERNAL (AT NODE) B.C.'s
    bc_flag = analysis.constr_bc_flag(nodes=nodes, constraints=constraints)

    # GENERATE STRIP WIDTH AND DIRECTION ANGLE
    el_props = analysis.elem_prop(nodes=nodes, elements=elements)

    # ENABLE cFSM ANALYSIS IF APPLICABLE, AND FIND BASE PROPERTIES
    if sum(gbt_con['glob']) + sum(gbt_con['dist']) \
                + sum(gbt_con['local']) + sum(gbt_con['other']) > 0:
        # turn on modal classification analysis
        cfsm_analysis = 1
        # set u_p stress to 1.0 for finding kg_global and k_global for axial modes
        nodes_base = deepcopy(nodes)
        nodes_base[:, 7] = np.ones_like(nodes[:, 7])

        # natural base first
        # properties all the longitudinal terms share
        [main_nodes, meta_elements, node_props, n_main_nodes, \
            n_corner_nodes, n_sub_nodes, n_dist_modes, n_local_modes, dof_perm] \
            = pycufsm.cfsm.base_properties(nodes=nodes_base, elements=elements)
        [r_x, r_z, r_yd, r_ys, r_ud] = pycufsm.cfsm.mode_constr(
            nodes=nodes_base,
            elements=elements,
            node_props=node_props,
            main_nodes=main_nodes,
            meta_elements=meta_elements
        )
        [d_y, n_global_modes] = pycufsm.cfsm.y_dofs(
            nodes=nodes_base,
            elements=elements,
            main_nodes=main_nodes,
            n_main_nodes=n_main_nodes,
            n_dist_modes=n_dist_modes,
            r_yd=r_yd,
            r_ud=r_ud,
            sect_props=sect_props,
            el_props=el_props
        )
    else:
        # no modal classification constraints are engaged
        cfsm_analysis = 0

    # LOOP OVER ALL THE LENGTHS TO BE INVESTIGATED
    for i, length in enumerate(lengths):
        # longitudinal terms to be included for this length
        m_a = m_all[i]

        total_m = len(m_a)  # Total number of longitudinal terms

        # SET SWITCH AND PREPARE BASE VECTORS (r_matrix) FOR cFSM ANALYSIS
        if cfsm_analysis == 1:
            # generate natural base vectors for axial compression loading
            b_v_l = pycufsm.cfsm.base_column(
                nodes_base=nodes_base,
                elements=elements,
                props=props,
                length=length,
                b_c=b_c,
                m_a=m_a,
                el_props=el_props,
                node_props=node_props,
                n_main_nodes=n_main_nodes,
                n_corner_nodes=n_corner_nodes,
                n_sub_nodes=n_sub_nodes,
                n_global_modes=n_global_modes,
                n_dist_modes=n_dist_modes,
                n_local_modes=n_local_modes,
                dof_perm=dof_perm,
                r_x=r_x,
                r_z=r_z,
                r_ys=r_ys,
                d_y=d_y
            )

        k_global, kg_global = analysis.k_kg_global(
            nodes=nodes,
            elements=elements,
            el_props=el_props,
            props=props,
            length=length,
            b_c=b_c,
            m_a=m_a
        )

        # %ADD SPRING CONTRIBUTIONS TO STIFFNESS
        # %Prior to version 4.3 this was the springs method
        #     %     if ~isempty(springs) %springs variable exists
        #     %         [k_global]=addspring(k_global,springs,n_nodes,length,b_c,m_a)
        #     %     end
        # %Now from version 4.3 this is the new springs method
        if len(springs) != 0:
            # springs variable exists
            for spring in springs:
                # Generate spring stiffness matrix (k_s) in local coordinates
                k_u = spring[3]
                k_v = spring[4]
                k_w = spring[5]
                k_q = spring[6]
                discrete = spring[8]
                y_s = spring[9] * length
                ks_l = analysis.spring_klocal(
                    k_u=k_u,
                    k_v=k_v,
                    k_w=k_w,
                    k_q=k_q,
                    length=length,
                    b_c=b_c,
                    m_a=m_a,
                    discrete=discrete,
                    y_s=y_s
                )

                # Transform k_s into global coordinates
                node_i = spring[1]
                node_j = spring[2]
                if node_j == 0 or spring[7] == 0:  # spring is to ground
                    # handle the spring to ground during assembly
                    alpha = 0  # use global coordinates for spring
                else:  # spring is between nodes
                    x_i = nodes[node_i, 1]
                    y_i = nodes[node_i, 2]
                    x_j = nodes[node_j, 1]
                    y_j = nodes[node_j, 2]
                    d_x = x_j - x_i
                    d_y = y_j - y_i
                    width = np.sqrt(d_x**2 + d_y**2)
                    if width < 1E-10:  # coincident nodes
                        alpha = 0  # use global coordinates for spring
                    else:
                        # local orientation for spring
                        alpha = np.arctan2(d_y, d_x)
                gamma = analysis.trans(alpha=alpha, total_m=total_m)
                k_s = gamma @ ks_l @ gamma.conj().T

                # Add element contribution of k_s to full matrix k_global
                k_global = analysis.spring_assemble(
                    k_global=k_global,
                    k_local=k_s,
                    node_i=node_i,
                    node_j=node_j,
                    n_nodes=n_nodes,
                    m_a=m_a
                )

        # INTERNAL BOUNDARY CONDITIONS (ON THE NODES) AND USER DEFINED CONSTR.
        # Check for user defined constraints too
        if bc_flag == 0:
            # no user defined constraints and fixities.
            r_u0_matrix = 0
            nu0 = 0
        else:
            # size boundary conditions and user constraints for use in r_matrix format
            # d_constrained=r_user*d_unconstrained, d=nodal DOF vector (note by
            # BWS June 5 2006)
            r_user = pycufsm.cfsm.constr_user(nodes=nodes, constraints=constraints, m_a=m_a)
            r_u0_matrix = spla.null_space(r_user.conj().T)
            # Number of boundary conditions and user defined constraints = nu0
            nu0 = len(r_u0_matrix[0])

        # %GENERATION OF cFSM CONSTRAINT MATRIX
        if cfsm_analysis == 1:
            # PERFORM ORTHOGONALIZATION IF GBT-LIKE MODES ARE ENFORCED
            b_v = pycufsm.cfsm.base_update(
                gbt_con=gbt_con,
                b_v_l=b_v_l,
                length=length,
                m_a=m_a,
                nodes=nodes,
                elements=elements,
                props=props,
                n_global_modes=n_global_modes,
                n_dist_modes=n_dist_modes,
                n_local_modes=n_local_modes,
                b_c=b_c,
                el_props=el_props
            )
            # no normalization is enforced: 0:  m
            # assign base vectors to constraints
            b_v = pycufsm.cfsm.mode_select(
                b_v=b_v,
                n_global_modes=n_global_modes,
                n_dist_modes=n_dist_modes,
                n_local_modes=n_local_modes,
                gbt_con=gbt_con,
                n_dof_m=4 * n_nodes,
                m_a=m_a
            )  # m
            r_mode = b_v
            # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        else:
            # no modal constraints are activated therefore
            r_mode = np.eye(4 * n_nodes * total_m)  # activate modal constraints

        # CREATE FINAL CONSTRAINT MATRIX
        # Determine the number of modal constraints, nm0
        if bc_flag == 0:
            # if no user defined constraints and fixities.
            r_matrix = r_mode
        else:
            # should performed uncoupled for block diagonal basis?
            if cfsm_analysis == 1:
                nm0 = 0
                r_m0_matrix = spla.null_space(r_mode.conj().T)
                nm0 = len(r_m0_matrix[0])
                r_0_matrix = r_m0_matrix
                if nu0 > 0:
                    r_0_matrix[:, nm0:(nm0 + nu0)] = r_u0_matrix
                r_matrix = spla.null_space(r_0_matrix.conj().T)
            else:
                r_matrix = spla.null_space(r_u0_matrix.conj().T)

        # INTRODUCE CONSTRAINTS AND REDUCE k_global MATRICES TO FREE PARTS ONLY
        k_global_ff = r_matrix.transpose() @ k_global @ r_matrix
        kg_global_ff = r_matrix.transpose() @ kg_global @ r_matrix
        # SOLVE THE EIGENVALUE PROBLEM
        # Determine which solver to use
        # small problems usually use eig (dense matrix),
        # and large problems use eigs (sparse matrix).
        # the eigs solver is not as stable as the full eig solver...
        # LAPACK reciprocal condition estimator
        # rcond_num = 1 / np.linalg.cond(np.linalg.pinv(kg_global_ff) @ k_global_ff)

        # Here, assume when rcond_num is bigger than half of the eps, eigs can provide
        # reliable solution. Otherwise, eig, the robust solver should be used.
        # if rcond_num >= np.spacing(1.0) / 2:
        #     eig_sparse = True
        #     # eigs
        # else:
        #     eig_sparse = False
        #     # eig

        # if eig_sparse:
        #     k_eigs = max(min(2*n_eigs, len(k_global_ff)), 1)
        #     if k_eigs == 1 or k_eigs == len(k_global_ff):
        #         [length_factors, modes] = spla.eig(
        #             a=k_global_ff,
        #             b=kg_global_ff
        #         )
        #     else:
        #         # pull out 10 eigenvalues
        #         [length_factors, modes] = eigs(
        #             A=k_global_ff,
        #             k=k_eigs,
        #             M=kg_global_ff,
        #             which='SM'
        #         )
        # else:
        [length_factors, modes] = spla.eig(a=k_global_ff, b=kg_global_ff)
        # CLEAN UP THE EIGEN SOLUTION
        # eigenvalues are along the diagonal of matrix length_factors
        #length_factors = np.diag(length_factors)
        # find all the positive eigenvalues and corresponding vectors, squeeze out the rest
        index = np.logical_and(length_factors > 0, abs(np.imag(length_factors)) < 0.00001)
        length_factors = length_factors[index]
        modes = modes[:, index]
        # sort from small to large
        index = np.argsort(length_factors)
        length_factors = length_factors[index]
        modes = modes[:, index]
        # only the real part is of interest
        # (eigensolver may give some small nonzero imaginary parts)
        length_factors = np.real(length_factors)
        modes = np.real(modes)

        # truncate down to reasonable number of modes to be kept
        num_pos_modes = len(length_factors)
        n_modes = min([n_eigs, num_pos_modes])
        length_factors = length_factors[:n_modes]
        modes = modes[:, :n_modes]

        # FORM THE FULL MODE SHAPE BY BRINGING BACK ELIMINATED DOF
        modes_full = r_matrix @ modes

        # CLEAN UP NORMALIZATION OF MODE SHAPE
        # eig and eigs solver use different normalization
        # set max entry (absolute) to +1.0 and scale the rest

        max_vals = np.amax(abs(modes_full), axis=0)
        for j in range(0, n_modes):
            modes_full[:, j] = modes_full[:, j] / max_vals[j]
        # GENERATE OUTPUT VALUES
        # curve and shapes are changed to cells!!
        # curve: buckling curve (load factor)
        # curve{i} = [mode# 1 ... mode#]
        # shapes = mode shapes
        # shapes{i} = mode, each column corresponds to a mode.
        if len(length_factors) > 0:
            signature[i] = length_factors[0]
        curve.append(length_factors)
        # shapes(:,i,1:min([n_modes,num_pos_modes]))=modes
        shapes.append(modes_full)

    return signature.flatten(), np.array(curve), np.array(shapes)



import numpy as np
from matplotlib.patches import Polygon
from matplotlib.collections import PatchCollection
from matplotlib.cm import jet  # pylint: disable=no-name-in-module
import matplotlib.pyplot as plt
import pycufsm.helpers as helpers


##Cross section
def crossect(node, elem, springs, constraint, flag):  # pylint: disable=unused-argument
    # BWS
    # October 2001 (last modified)
    # December 2015 springs updated
    # December 2015 origin updated to also show axes
    # December 2015 added stress distribution into this plot instead of separate
    # December 2015 added an additional flag for section property axes
    # plots the cross-section
    #
    # node: [node# x z dofx dofz dofy dofrot stress] nnodes x 8;
    # elem: [elem# nodei nodej t mat#] nelems x 4;
    # flags:[node# element# mat# stress# stresspic coord constraints
    #        springs origin propaxis] 1 means show
    nodeflag = flag[0]
    elemflag = flag[1]
    matflag = flag[2]
    stressflag = flag[3]
    stresspicflag = flag[4]
    # coordflag = flag[5]
    constraintsflag = flag[6]
    # springsflag = flag[7]
    originflag = flag[8]
    patches = []
    # if len(flag) > 10:
    #     propaxisflag = flag[9]
    # else:
    #     propaxisflag = 0
    if stresspicflag == 1:
        scale = 1
        maxstress = max(np.abs(node[:, 7]))
        stress = np.append(
            node[:, 0].reshape((len(node), 1)), (node[:, 7] / maxstress).reshape((len(node), 1)),
            axis=1
        )
        maxi = np.max(np.abs(node[:, 1:3]))
        maxoffset = scale * np.max(maxi) / 10
        stresscord = np.zeros((len(node), 3))
        for i in range(len(stress)):
            stresscord[i, 0:3] = [
                node[i, 0], node[i, 1] + maxoffset * stress[i, 1],
                node[i, 2] - maxoffset * stress[i, 1]
            ]
    #Plot the nodes
    _, ax1 = plt.subplots(constrained_layout=True, figsize=(6, 6))
    plt.plot(node[:, 1], node[:, 2], 'bo', markersize=2)
    #Plot the elements
    for i in range((len(elem))):
        nodei = int(elem[i, 1])
        nodej = int(elem[i, 2])
        x_i = node[nodei, 1]
        z_i = node[nodei, 2]
        x_j = node[nodej, 1]
        z_j = node[nodej, 2]
        theta = np.arctan2((z_j - z_i), (x_j - x_i))
        thick = elem[i, 3] * 1
        points = np.array([[x_i - np.sin(theta) * thick / 2, z_i + np.cos(theta) * thick / 2],
                           [x_j - np.sin(theta) * thick / 2, z_j + np.cos(theta) * thick / 2],
                           [x_j + np.sin(theta) * thick / 2, z_j - np.cos(theta) * thick / 2],
                           [x_i + np.sin(theta) * thick / 2, z_i - np.cos(theta) * thick / 2]])
        plt.plot([x_i, x_j], [z_i, z_j], 'bo', markersize=0.5)
        polygon = Polygon(points, True, ec='b', fc=(1, 1, 0, 1), lw=0.5)
        ax1.add_artist(polygon)
        #patches.append(polygon)
        if stresspicflag == 1:
            #get the stresses
            sxi = stresscord[nodei, 1]
            szi = stresscord[nodei, 2]
            sxj = stresscord[nodej, 1]
            szj = stresscord[nodej, 2]
            #plot the stress in pseudo 3D
            if node[nodei, 7] >= 0:
                plt.plot([x_i, sxi], [z_i, szi], 'r')
            else:
                plt.plot([x_i, sxi], [z_i, szi], 'b')
            if node[nodej, 7] >= 0:
                plt.plot([x_j, sxj], [z_j, szj], 'r')
            else:
                plt.plot([x_j, sxj], [z_j, szj], 'b')
            plt.plot([sxi, sxj], [szi, szj], 'k')
            if stressflag == 1:
                plt.text(sxi, szi, str(round(node[nodei, 7], 2)))
                plt.text(sxj, szj, str(round(node[nodej, 7], 2)))
        #plot the element labels if wanted
        if elemflag == 1:
            plt.text((x_i+x_j) / 2, (z_i+z_j) / 2, str(elem[i, 0] + 1), fontsize=8)
        #plot the materials labels if wanted
        if matflag == 1:
            plt.text((x_i+x_j) / 2 + 10, (z_i+z_j) / 2 + 10, str(elem[i, 4]), fontsize=8)
        #Plot th stress distribution in 3D if wanted
        #####___#####
    ####Patches of cross section
    PatchCollection(patches, cmap=jet, alpha=0.4)
    #colors = np.zeros(len(patches))
    #patch.set_array(np.array(colors))
    #plt.add_collection(patch)
    #plt.xlim((x_min - 25, x_max + 25))
    #plt.ylim((y_min - 25, y_max + 25))
    #Plot the node labels if wanted
    if nodeflag == 1:
        for i in range(len(node)):
            plt.text(node[i, 1], node[i, 2], str(node[i, 0] + 1))
    #Plot the stress at the node if wanted
    if stressflag == 1 and stresspicflag == 0:
        for i in range(len(node)):
            plt.text(node[i, 1], node[i, 2], str(round(node[i, 7], 2)))
    #Plot the origin point
    if originflag == 1:
        plt.plot(
            0,
            0,
            'ko',
        )
        xmax = np.max(np.max(node[:, 1]))
        zmax = np.max(np.max([node[:, 2]]))
        ax_len = min(xmax, zmax)
        plt.plot([0, 0.2 * ax_len], [0, 0], 'k')
        plt.text(0.22 * ax_len, 0, 'x_o')
        plt.plot([0, 0], [0, 0.2 * ax_len], 'k')
        plt.text(0, 0.22 * ax_len, 'z_o')
    if constraintsflag == 1:
        for i in range(len(node)):
            dofx = node[i, 3]
            dofz = node[i, 4]
            dofy = node[i, 5]
            dofq = node[i, 6]
            if min([dofx, dofz, dofy, dofq]) == 0:
                plt.plot(node[i, 1], node[i, 2], 'sq')
        if len(constraint) == 0:
            print('No constraints')
        else:
            for i in range(len(constraint)):
                nodee = constraint[i, 0]
                nodek = constraint[i, 3]
                plt.plot(node[nodee, 1], node[nodee, 2], 'xg')
                plt.plot(node[nodek, 1], node[nodek, 2], 'hg')
    #Plot the springs if wanted
    ####SPRINGS AND CONSTRAINTS REMAINING
    #springsscale = 0.05*np.max(np.max(np.abs(node[:, 1:3])))
    plt.gca().set_aspect('equal', adjustable='box')
    plt.axis('off')
    # plt.savefig('Validation/'+address+'/CS.png')
    plt.show()


#Cross section displacement function
def dispshap(undef, node, elem, mode, scalem, springs, m_a, b_c, surf_pos):  # pylint: disable=unused-argument
    #Determining Scaling Factor for the displaced shape
    ##dispmax=np.max(np.abs(mode))
    dispmax = np.max(np.abs(mode))
    membersize = np.max(np.max(node[:, 1:2])) - np.min(np.min(node[:, 1:2]))
    scale = scalem * membersize / dispmax / 10
    #Generate and Plot
    _, axes = plt.subplots(constrained_layout=True, figsize=(6, 6))
    x_max = -np.inf
    y_max = -np.inf
    x_min = np.inf
    y_min = np.inf
    defpoints = []
    if undef == 1:
        for i in range(len(elem)):
            nodei = int(elem[i, 1])
            nodej = int(elem[i, 2])
            x_i = node[nodei, 1]
            x_j = node[nodej, 1]
            z_i = node[nodei, 2]
            z_j = node[nodej, 2]
            #PLOT undeformed geometry
            theta = np.arctan2((z_j - z_i), (x_j - x_i))
            thick = elem[i, 3]
            points = np.array([[x_i - np.sin(theta) * thick / 2, z_i + np.cos(theta) * thick / 2],
                               [x_j - np.sin(theta) * thick / 2, z_j + np.cos(theta) * thick / 2],
                               [x_j + np.sin(theta) * thick / 2, z_j - np.cos(theta) * thick / 2],
                               [x_i + np.sin(theta) * thick / 2, z_i - np.cos(theta) * thick / 2]])
            #Plot axis limits
            x_max = max(x_max, np.max(points[:, 0]))
            y_max = max(y_max, np.max(points[:, 1]))
            x_min = min(x_min, np.min(points[:, 0]))
            y_min = min(y_min, np.min(points[:, 1]))
            #points = np.random.rand(5 ,2)
            polygon = Polygon(points, True, ec='b', fc='y', lw=0.5)
            axes.add_artist(polygon)
            plt.plot([x_i, x_j], [z_i, z_j], 'bo', markersize=2)
    #patch = PatchCollection(patches, cmap =jet, alpha=0.4)
    # colors = np.zeros(len(patches))
    # patch.set_array(np.array(colors))
    #axes.add_collection(patch)
    #plt.xlim((x_min - 25, x_max + 25))
    #plt.ylim((y_min - 25, y_max + 25))
    nnodes = len(node)
    for i in range(len(elem)):
        #Get Element Geometry
        nodei = int(elem[i, 1])
        nodej = int(elem[i, 2])
        x_i = node[nodei, 1]
        x_j = node[nodej, 1]
        z_i = node[nodei, 2]
        z_j = node[nodej, 2]
        #Determine the global element displacements
        #dbar is the nodal displacements for the element in global
        #coordinates dbar=[u1 v1 u2 v2 w1 o1 w2 o2]
        dbar = np.zeros((8, 1))
        dbarm = np.zeros((8, 1))
        dlbarm = np.zeros((3, 9))
        defpatches = []
        for j, m_a_j in enumerate(m_a):
            dbar[0:2, 0] = mode[4*nnodes*j + 2 * (nodei+1) - 2:4*nnodes*j + 2 * (nodei+1)]
            dbar[2:4, 0] = mode[4*nnodes*j + 2 * (nodej+1) - 2:4*nnodes*j + 2 * (nodej+1)]
            dbar[4:6, 0] = mode[4*nnodes*j + 2*nnodes + 2 * (nodei+1) - 2:4*nnodes*j + 2*nnodes
                                + 2 * (nodei+1)]
            dbar[6:8, 0] = mode[4*nnodes*j + 2*nnodes + 2 * (nodej+1) - 2:4*nnodes*j + 2*nnodes
                                + 2 * (nodej+1)]
            #Transform dbar into local coordinates
            phi = np.arctan2(-(z_j - z_i), (x_j - x_i))
            disp = helpers.gammait(phi, dbar)
            #Determine additional displacements in each element
            links = 10
            length = np.sqrt((x_j - x_i)**2 + (z_j - z_i)**2)
            disp_local = helpers.shapef(links, disp, length)
            #Transform additional displacements into global coordinates
            dlbar = helpers.gammait2(phi, disp_local)
            cutloc = 1 / surf_pos
            if b_c.startswith('S-S'):
                dbarm = dbar * np.sin(m_a_j * np.pi / cutloc) + dbarm
                dlbarm = dlbar * np.sin(m_a_j * np.pi / cutloc) + dlbarm
            elif b_c.startswith('C-C'):
                dbarm = dbar * np.sin(m_a_j * np.pi / cutloc) * np.sin(np.pi / cutloc) + dbarm
                dlbarm = dlbar * np.sin(m_a_j * np.pi / cutloc) * np.sin(np.pi / cutloc) + dlbarm
            elif b_c.startswith('S-C') or b_c.startswith('C-S'):
                dbarm = dbar * (
                    np.sin((m_a_j+1) * np.pi / cutloc) + (m_a_j+1) * np.sin(np.pi / cutloc) / m_a_j
                ) + dbarm
                dlbarm = dlbar * (
                    np.sin((m_a_j+1) * np.pi / cutloc) + (m_a_j+1) * np.sin(np.pi / cutloc) / m_a_j
                ) + dlbarm
            elif b_c.startswith('F-C') or b_c.startswith('C-F'):
                dbarm = dbar * (1 - np.cos((m_a_j - 1/2) * np.pi / cutloc)) + dbarm
                dlbarm = dlbar * (1 - np.cos((m_a_j - 1/2) * np.pi / cutloc)) + dlbarm
            elif b_c.startswith('G-C') or b_c.startswith('C-G'):
                dbarm = dbar * (
                    np.sin((m_a_j - 1/2) * np.pi / cutloc) * np.sin(np.pi / cutloc / 2)
                ) + dbarm
                dlbarm = dlbar * (
                    np.sin((m_a_j - 1/2) * np.pi / cutloc) * np.sin(np.pi / cutloc / 2)
                ) + dlbarm
        #Create a vertor of undisplaced coordinates "undisp"
        undisp = np.zeros((2, links + 1))
        # undisp[:, 0] = np.transpose([x_i, z_i])
        # undisp[:, links] = np.transpose([x_j, z_j])
        for j in range(0, links + 1):
            undisp[:,
                   j] = np.transpose([x_i + (x_j-x_i) * (j) / links, z_i + (z_j-z_i) * (j) / links])
        #create a vector of displaced coordinated "disp"
        disp = np.zeros((2, links + 1))
        disp[:, 0] = np.transpose([x_i + scale * dbarm[0], z_i + scale * dbarm[4]])
        disp[:, links] = np.transpose([x_j + scale * dbarm[2], z_j + scale * dbarm[6]])
        disp[0, 1:links] = undisp[0, 1:links] + scale * dlbarm[0, :]
        disp[1, 1:links] = undisp[1, 1:links] + scale * dlbarm[2, :]
        #The angle of each link
        thetalinks = np.arctan2(
            disp[1, 1:links + 1] - disp[1, 0:links], disp[0, 1:links + 1] - disp[0, 0:links]
        )
        thetalinks = np.append(thetalinks, thetalinks[links - 1])
        #Plot the deformed geometry
        theta = np.arctan2((z_j - z_i), (x_j - x_i))
        thick = elem[i, 3]
        #Deformed geomtery with appropriate thickness
        dispout = np.array([[disp[0, :] + np.sin(thetalinks) * thick / 2],
                            [disp[1, :] - np.cos(thetalinks) * thick / 2]]).T
        dispin = np.array([[disp[0, :] - np.sin(thetalinks) * thick / 2],
                           [disp[1, :] + np.cos(thetalinks) * thick / 2]]).T
        dispout = dispout.reshape((11, 2))
        dispin = dispin.reshape((11, 2))
        for j in range(links):
            defpoints = np.array([[dispout[j, 0], dispout[j, 1]], [dispin[j, 0], dispin[j, 1]],
                                  [dispin[j + 1, 0], dispin[j + 1, 1]],
                                  [dispout[j + 1, 0], dispout[j + 1, 1]]])
            polygon = Polygon(defpoints, True, ec='r', fc='r', lw=0.5)
            #defpatches = defpatches.append(polygon)
            axes.add_artist(polygon)
        plt.plot([disp[0, 0], disp[0, links]], [disp[1, 0], disp[1, links]], 'bo', markersize=2)
    PatchCollection(defpatches, cmap=jet, alpha=0.4)
    # dcolors = 100*np.random.rand(len(patches))
    # def_patch.set_array(np.array(dcolors))
    #axes.add_collection(def_patch)
    plt.gca().set_aspect('equal', adjustable='box')
    plt.axis('off')


    # if(figure == 1):
    #     plt.savefig('Validation/'+address+'/local.png')
    # if(figure == 2):
    #     plt.savefig('Validation/'+address+'/distortional.png')
    # if(figure == 3):
    #     plt.savefig('Validation/'+address+'/global.png')
    # if(figure == 4):
    #     plt.savefig('Validation/'+address+'/global1.png')
    # plt.show()
def thecurve3(
    curvecell, clas, filedisplay, minopt, logopt, clasopt, xmin, xmax, ymin, ymax, modedisplay,
    fileindex, modeindex, picpoint
):  # pylint: disable=unused-argument
    curve = curvecell
    marker = '.x+*sdv^<'
    color1 = 'bgky'
    _, ax2 = plt.subplots(constrained_layout=True, figsize=(6, 6))
    for file_i in filedisplay:
        mark = ['b', marker[(file_i) % 10]]
        if logopt == 1:
            for j, mode in enumerate(modedisplay):
                ax2.semilogx(
                    curve[:, mode - 1, 0],
                    curve[:, mode - 1, 1],
                    color=color1[(j % 4)],
                    marker=mark[1]
                )
                # ax2.semilogx(curve_sign[:,0], curve_sign[:,1], 'k')
        else:
            for mode in modedisplay:
                ax2.plot(
                    curve[:, mode - 1, 0], curve[:, mode - 1, 1], color=mark[0], marker=mark[1]
                )
                # ax2.plot(curve_sign[:,0], curve_sign[:,1], 'k')

        c_r = 0
        if minopt == 1:
            for mode in modedisplay:
                for j in range(len(curve[:, 1, 1]) - 2):
                    load1 = curve[j, mode - 1, 1]
                    load2 = curve[j + 1, mode - 1, 1]
                    load3 = curve[j + 2, mode - 1, 1]
                    if load2 < load1 and load2 <= load3:
                        c_r = c_r + 1
                        ax2.text(
                            curve[j + 1, mode - 1, 0], curve[j + 1, mode - 1, 1] - (ymax-ymin) / 20,
                            f"{curve[j + 1, mode - 1, 0]:.2f}, {curve[j + 1, mode - 1, 1]:.2f}"
                        )
    # ax2.text(picpoint[0], picpoint[1],
    #  "{0:.2f}, {0:.2f}".format(curve[j + 1, j, 0], curve[j + 1, j, 1], color = 'r' )
    # )
    plt.xlim((xmin, xmax))
    plt.ylim((ymin, ymax))
    plt.xlabel('length')
    plt.ylabel('load factor')
    plt.title('Buckling curve')
    plt.show()
    #set the callback of curve



import numpy as np
import scipy.io
from pytest import approx, raises
from pycufsm import cutwp, preprocess, helpers, fsm
import pycufsm.examples.example_1 as example_1
from .utils import pspec_context
from .fixtures.e2e_fixtures import *


def mat_file_test(mat_filename):
    mat = scipy.io.loadmat("tests/mat_files/" + mat_filename)
    cufsm_input = helpers.load_mat(mat)

    # Disable cFSM for now in basic loading of mat files
    cufsm_input["GBTcon"]["glob"] = [0]
    cufsm_input["GBTcon"]["dist"] = [0]
    cufsm_input["GBTcon"]["local"] = [0]
    cufsm_input["GBTcon"]["other"] = [0]

    coords = cufsm_input["nodes"][:, 1:3]
    ends = cufsm_input["elements"][:, 1:4]
    sect_props_cutwp = cutwp.prop2(coords, ends)
    m_all_ones = np.ones((len(cufsm_input["lengths"]), 1))
    signature, curve, shapes = fsm.strip(
        props=cufsm_input["props"],
        nodes=cufsm_input["nodes"],
        elements=cufsm_input["elements"],
        lengths=cufsm_input["lengths"],
        springs=cufsm_input["springs"],
        constraints=cufsm_input["constraints"],
        gbt_con=cufsm_input["GBTcon"],
        b_c='S-S',
        m_all=m_all_ones,
        n_eigs=1,
        sect_props=sect_props_cutwp
    )
    lengths = list(cufsm_input["lengths"])
    return lengths, mat["curve"], sect_props_cutwp, signature, curve, shapes


def describe_end_to_end_tests():

    @pspec_context("End-to-End Tests")
    def describe():
        pass

    def context_example_1():

        @pspec_context("Example 1")
        def describe():
            pass

        results = example_1.__main__()

        def it_results_in_correct_solution():
            assert results["X_values"] == [
                0.5, 0.75, 1, 1.25, 1.5, 1.75, 2, 2.25, 2.5, 2.75, 3, 3.25, 3.5, 3.75, 4, 4.25, 4.5,
                4.75, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 22, 24, 26, 28, 30,
                32, 34, 36, 38, 40, 42, 44, 46, 48, 50, 52, 54, 56, 58, 60, 66, 72, 78, 84, 90, 96,
                102, 108, 114, 120, 132, 144, 156, 168, 180, 204, 228, 252, 276, 300
            ]

    def context_dsm_3_2_1_c_with_lips_Mx():

        @pspec_context("DSM Guide Jan 2006, Ex 3.2.1: C-Section with Lips (Mx)")
        def describe():
            pass

        lengths, expected_curve, sect_props, signature, curve, _ = mat_file_test("cwlip_Mx.mat")

        expected = {"xcrl": 5, "Mcrl": 0.67, "xcrd": 26.6, "Mcrd": 0.85, "A": 0.880, "Ixx": 10.285}

        def ir_results_in_correct_sect_props():
            assert expected["A"] == approx(sect_props["A"], abs=0.001)
            assert expected["Ixx"] == approx(sect_props["Ixx"], abs=0.001)

        def it_results_in_correct_Mcr():
            assert signature[lengths.index(expected["xcrl"])] == approx(expected["Mcrl"], abs=0.01)
            assert signature[lengths.index(expected["xcrd"])] == approx(expected["Mcrd"], abs=0.01)

        def it_results_in_correct_signature_curve():
            assert np.allclose(expected_curve[:, 1, 0], curve[:, 0], rtol=1.e-4)

    def context_dsm_3_2_1_c_with_lips_P():

        @pspec_context("DSM Guide Jan 2006, Ex 3.2.1: C-Section with Lips (P)")
        def describe():
            pass

        lengths, expected_curve, sect_props, signature, curve, _ = mat_file_test("cwlip_P.mat")

        expected = {
            "xcrl": 6.6,
            "Pcrl": 0.12,
            "xcrd": 28.5,
            "Pcrd": 0.27,
            "A": 0.880,
            "Ixx": 10.285
        }

        def ir_results_in_correct_sect_props():
            assert expected["A"] == approx(sect_props["A"], abs=0.001)
            assert expected["Ixx"] == approx(sect_props["Ixx"], abs=0.001)

        def it_results_in_correct_Pcr():
            assert signature[lengths.index(expected["xcrl"])] == approx(expected["Pcrl"], abs=0.01)
            assert signature[lengths.index(expected["xcrd"])] == approx(expected["Pcrd"], abs=0.01)

        def it_results_in_correct_signature_curve():
            assert np.allclose(expected_curve[:, 1, 0], curve[:, 0], rtol=1.e-4)

    def context_dsm_3_2_2_c_with_lips_modified_Mx():

        @pspec_context("DSM Guide Jan 2006, Ex 3.2.2: Modified C-Section with Lips (Mx)")
        def describe():
            pass

        lengths, expected_curve, sect_props, signature, curve, _ = mat_file_test(
            "cwlip_modified_Mx.mat"
        )

        expected = {
            "xcrl": 2.7,
            "Mcrl": 1.40,
            "xcrd": 30.5,
            "Mcrd": 0.98,
            "A": 0.933,
            "Ixx": 10.818
        }

        def ir_results_in_correct_sect_props():
            assert expected["A"] == approx(sect_props["A"], abs=0.001)
            assert expected["Ixx"] == approx(sect_props["Ixx"], abs=0.001)

        def it_results_in_correct_Mcr():
            assert signature[lengths.index(expected["xcrl"])] == approx(expected["Mcrl"], abs=0.01)
            assert signature[lengths.index(expected["xcrd"])] == approx(expected["Mcrd"], abs=0.01)

        def it_results_in_correct_signature_curve():
            assert np.allclose(expected_curve[:, 1, 0], curve[:, 0], rtol=1.e-3)

    def context_dsm_3_2_2_c_with_lips_modified_P():

        @pspec_context("DSM Guide Jan 2006, Ex 3.2.2: Modified C-Section with Lips (P)")
        def describe():
            pass

        lengths, expected_curve, sect_props, signature, curve, _ = mat_file_test(
            "cwlip_modified_P.mat"
        )

        expected = {
            "xcrl": 11.5,
            "Pcrl": 0.27,
            "xcrd": 32.7,
            "Pcrd": 0.32,
            "A": 0.933,
            "Ixx": 10.818
        }

        def ir_results_in_correct_sect_props():
            assert expected["A"] == approx(sect_props["A"], abs=0.001)
            assert expected["Ixx"] == approx(sect_props["Ixx"], abs=0.001)

        def it_results_in_correct_Pcr():
            assert signature[lengths.index(expected["xcrl"])] == approx(expected["Pcrl"], abs=0.01)
            assert signature[lengths.index(expected["xcrd"])] == approx(expected["Pcrd"], abs=0.01)

        def it_results_in_correct_signature_curve():
            assert np.allclose(expected_curve[:, 1, 0], curve[:, 0], rtol=1.e-4)

    def context_dsm_3_2_5_z_with_lips_Mx():

        @pspec_context("DSM Guide Jan 2006, Ex 3.2.2: Z-Section with Lips (Mx)")
        def describe():
            pass

        lengths, expected_curve, sect_props, signature, curve, _ = mat_file_test("zwlip_Mxr.mat")

        expected = {"xcrl": 4.1, "Mcrl": 0.85, "xcrd": 22.1, "Mcrd": 0.77, "A": 0.822, "Ixx": 7.762}

        def ir_results_in_correct_sect_props():
            assert expected["A"] == approx(sect_props["A"], abs=0.001)
            assert expected["Ixx"] == approx(sect_props["Ixx"], abs=0.001)

        def it_results_in_correct_Mcr():
            assert signature[lengths.index(expected["xcrl"])] == approx(expected["Mcrl"], abs=0.01)
            assert signature[lengths.index(expected["xcrd"])] == approx(expected["Mcrd"], abs=0.01)

        def it_results_in_correct_signature_curve():
            # TODO: Investigate why the signature curve starts mismatching at the tail end...
            assert np.allclose(expected_curve[0:-10, 1, 0], curve[0:-10, 0], rtol=1.e-3)

    def context_dsm_3_2_5_z_with_lips_P():

        @pspec_context("DSM Guide Jan 2006, Ex 3.2.2: Z-Section with Lips (P)")
        def describe():
            pass

        lengths, expected_curve, sect_props, signature, curve, _ = mat_file_test("zwlip_P.mat")

        expected = {"xcrl": 5.9, "Pcrl": 0.16, "xcrd": 18.3, "Pcrd": 0.29, "A": 0.822, "Ixx": 7.762}

        def ir_results_in_correct_sect_props():
            assert expected["A"] == approx(sect_props["A"], abs=0.001)
            assert expected["Ixx"] == approx(sect_props["Ixx"], abs=0.001)

        def it_results_in_correct_Pcr():
            assert signature[lengths.index(expected["xcrl"])] == approx(expected["Pcrl"], abs=0.01)
            assert signature[lengths.index(expected["xcrd"])] == approx(expected["Pcrd"], abs=0.01)

        def it_results_in_correct_signature_curve():
            # TODO: Investigate why the signature curve starts mismatching at the tail end...
            assert np.allclose(expected_curve[0:-10, 1, 0], curve[0:-10, 0], rtol=1.e-3)



import numpy as np
from scipy import linalg as spla

# Originally developed for MATLAB by Benjamin Schafer PhD et al
# Ported to Python by Brooks Smith MEng, PE
#
# Each function within this file was originally its own separate file.
# Original MATLAB comments, especially those retaining to authorship or
# change history, have been generally retained unaltered


def template_path(draw_table, thick, n_r=4, shift=None):
    # Brooks H. Smith
    # 17 June 2020
    # Assuming a uniform thickness, draws a section according to a path definition
    # draw_table = matrix of the form [[theta, dist, rad, n_s]], where:
    #              theta = starting angle, dist = length of straight segment,
    #              rad = radius of curved segment, n_s = number of mesh elements in straight
    # thick = thickness
    # n_r = number of mesh elements in curved segments
    if shift is None:
        shift = [0, 0]

    nodes = []
    elements = []
    if len(draw_table[0]) == 5:
        n_r_table = True
    else:
        n_r_table = False

    # Set initial point
    if draw_table[0][1] != 0:
        nodes.append(np.array([len(nodes), shift[0], shift[1], 1, 1, 1, 1, 1.0]))

    # Progress through drawing the section
    for i, row in enumerate(draw_table[:-1]):
        theta = row[0]
        dist = row[1]
        rad = row[2]
        n_s = row[3]
        next_theta = draw_table[i + 1][0]
        phi = np.mod(next_theta - theta, 2 * np.pi)
        if phi > np.pi:
            phi = phi - 2 * np.pi

        # Add elements in straight segment (if n_s > 1)
        for j in range(1, int(n_s)):
            x_loc = nodes[-1][1] + dist / n_s * np.cos(theta)
            y_loc = nodes[-1][2] + dist / n_s * np.sin(theta)
            nodes.append(np.array([len(nodes), x_loc, y_loc, 1, 1, 1, 1, 1.0]))

        # Add elements in curved segment
        centre = [
            nodes[-1][1] + dist / n_s * np.cos(theta) - np.sign(phi) * rad * np.sin(theta),
            nodes[-1][2] + dist / n_s * np.sin(theta) + np.sign(phi) * rad * np.cos(theta),
        ]
        if n_r_table:
            n_r = row[4]

        if rad == 0:
            nodes.append(np.array([len(nodes), centre[0], centre[1], 1, 1, 1, 1, 1.0]))
        else:
            for j in range(int(n_r)):
                theta_j = theta - np.sign(phi) * np.pi / 2 + j * 1.0 / max(1, n_r - 1) * phi
                x_loc = centre[0] + rad * np.cos(theta_j)
                y_loc = centre[1] + rad * np.sin(theta_j)
                nodes.append(np.array([len(nodes), x_loc, y_loc, 1, 1, 1, 1, 1.0]))

    # Draw the last straight line
    theta = draw_table[-1][0]
    dist = draw_table[-1][1]
    n_s = draw_table[-1][3]
    if dist > 0:
        for j in range(1, int(n_s) + 1):
            x_loc = nodes[-1][1] + dist / n_s * np.cos(theta)
            y_loc = nodes[-1][2] + dist / n_s * np.sin(theta)
            nodes.append(np.array([len(nodes), x_loc, y_loc, 1, 1, 1, 1, 1.0]))

    # build the elements list
    for i in range(0, len(nodes) - 1):
        elements.append(np.array([i, i, i + 1, thick, 0]))

    return [np.array(nodes), np.array(elements)]


def template_calc(sect):
    n_d = sect['n_d']
    n_b1 = sect['n_b1']
    n_b2 = sect['n_b2']
    n_l1 = sect['n_l1']
    n_l2 = sect['n_l2']
    n_r = sect['n_r']

    # BWS
    # August 23, 2000
    # 2015 modification to allow for l_1=l_2=0 and creation of a track with same template
    # 2015 addition to allow outer dimensions and inner radii to be used
    # 2015 addition to control element discretization

    nodes = []
    elements = []

    # CorZ=determines sign conventions for flange 1=C 2=Z
    if sect['type'] == 'Z':
        flip_b2 = -1
    else:
        flip_b2 = 1

    # channel template
    # convert angles to radians
    q_1 = 90 * np.pi / 180
    q_2 = 90 * np.pi / 180

    # outer dimensions and inner radii came in and these
    # need to be corrected to all centerline for the use of this template
    [depth, b_1, l_1, b_2, l_2, rad, thick] = template_out_to_in(sect)

    # rest of the dimensions are "flat dimensions" and acceptable for modeling
    if rad == 0:
        if l_1 == 0 and l_2 == 0:
            # track or unlipped Z with sharp corners
            geom = [{
                'x': b_1,
                'y': 0,
                'n': n_b1,
                'r': False
            }, {
                'x': 0,
                'y': 0,
                'n': n_d,
                'r': False
            }, {
                'x': 0,
                'y': depth,
                'n': n_b2,
                'r': False
            }, {
                'x': flip_b2 * (b_2),
                'y': depth,
                'n': 0,
                'r': False
            }]
        else:
            # lipped C or Z with sharp corners
            geom = [{
                'x': b_1 + l_1 * np.cos(q_1),
                'y': l_1 * np.sin(q_1),
                'n': n_l1,
                'r': False
            }, {
                'x': b_1,
                'y': 0,
                'n': n_b1,
                'r': False
            }, {
                'x': 0,
                'y': 0,
                'n': n_d,
                'r': False
            }, {
                'x': 0,
                'y': depth,
                'n': n_b2,
                'r': False
            }, {
                'x': flip_b2 * (b_2),
                'y': depth,
                'n': n_l2,
                'r': False
            }, {
                'x': flip_b2 * (b_2 + l_2 * np.cos(q_2)),
                'y': depth - l_2 * np.sin(q_2),
                'n': 0,
                'r': False
            }]

    else:
        # Unlipped C or Z with round corners
        if l_1 == 0 and l_2 == 0:
            geom = [{
                'x': rad + b_1,
                'y': 0,
                'n': n_b1,
                'r': False
            }, {
                'x': rad,
                'y': 0,
                'n': n_r,
                'r': True
            }, {
                'x': 0,
                'y': rad,
                'n': n_d,
                'r': False
            }, {
                'x': 0,
                'y': rad + depth,
                'n': n_r,
                'r': True
            }, {
                'x': flip_b2 * rad,
                'y': rad + depth + rad,
                'n': n_b2,
                'r': False
            }, {
                'x': flip_b2 * (rad+b_2),
                'y': rad + depth + rad,
                'n': 0,
                'r': False
            }]
        # lipped C or Z with round corners
        else:
            geom = [{
                'x': rad + b_1 + rad * np.cos(np.pi / 2 - q_1) + l_1 * np.cos(q_1),
                'y': rad - rad * np.sin(np.pi / 2 - q_1) + l_1 * np.sin(q_1),
                'n': n_l1,
                'r': False
            }, {
                'x': rad + b_1 + rad * np.cos(np.pi / 2 - q_1),
                'y': rad - rad * np.sin(np.pi / 2 - q_1),
                'n': n_r,
                'r': True
            }, {
                'x': rad + b_1,
                'y': 0,
                'n': n_b1,
                'r': False
            }, {
                'x': rad,
                'y': 0,
                'n': n_r,
                'r': True
            }, {
                'x': 0,
                'y': rad,
                'n': n_d,
                'r': False
            }, {
                'x': 0,
                'y': rad + depth,
                'n': n_r,
                'r': True
            }, {
                'x': flip_b2 * rad,
                'y': rad + depth + rad,
                'n': n_b2,
                'r': False
            }, {
                'x': flip_b2 * (rad+b_2),
                'y': rad + depth + rad,
                'n': n_r,
                'r': True
            }, {
                'x': flip_b2 * (rad + b_2 + rad * np.cos(np.pi / 2 - q_2)),
                'y': rad + depth + rad - rad + rad * np.sin(np.pi / 2 - q_2),
                'n': n_l2,
                'r': False
            }, {
                'x': flip_b2 * (rad + b_2 + rad * np.cos(np.pi / 2 - q_2) + l_2 * np.cos(q_2)),
                'y': rad + depth + rad - rad + rad * np.sin(np.pi / 2 - q_2) - l_2 * np.sin(q_2),
                'n': 0,
                'r': False
            }]

    # mesh it
    # number of elements between the geom coordinates
    for i, (g_1, g_2) in enumerate(zip(geom[:-1], geom[1:])):
        d_x = g_2['x'] - g_1['x']
        d_y = g_2['y'] - g_1['y']
        # nodes.append(
        #    np.array([len(nodes), g_1['x'], g_1['y'], 1, 1, 1, 1, 1.0]))

        if g_1['r']:
            # ROUND CORNER MODEL
            if l_1 == 0 and l_2 == 0:
                # ------------------------
                # UNLIPPED C OR Z SECTION
                for j in range(0, g_1['n']):
                    if i == 1:
                        x_c = rad
                        y_c = rad
                        q_start = np.pi / 2
                        d_q = np.pi / 2 * j / g_1['n']
                    if i == 3:
                        x_c = flip_b2 * rad
                        y_c = rad + depth
                        q_start = np.pi if flip_b2 == 1 else 0
                        d_q = flip_b2 * np.pi / 2 * j / g_1['n']
                    x_2 = x_c + rad * np.cos(q_start + d_q)
                    # note sign on 2nd term is negative due to y sign convention (down positive)
                    y_2 = y_c - rad * np.sin(q_start + d_q)
                    nodes.append(np.array([len(nodes), x_2, y_2, 1, 1, 1, 1, 1.0]))
                # ------------------------
            else:
                # ------------------------
                # LIPPED C OR Z SECTION
                # we are in a corner and must be fancier
                for j in range(0, g_1['n']):
                    if i == 1:
                        x_c = rad + b_1
                        y_c = rad
                        q_start = np.pi / 2 - q_1
                        d_q = q_1 * j / g_1['n']
                    if i == 3:
                        x_c = rad
                        y_c = rad
                        q_start = np.pi / 2
                        d_q = np.pi / 2 * j / g_1['n']
                    if i == 5:
                        x_c = flip_b2 * rad
                        y_c = rad + depth
                        q_start = np.pi if flip_b2 == 1 else 0
                        d_q = flip_b2 * np.pi / 2 * j / g_1['n']
                    if i == 7:
                        x_c = flip_b2 * (rad+b_2)
                        y_c = rad + depth + rad - rad
                        q_start = 3 * np.pi / 2
                        d_q = flip_b2 * q_2 * j / g_1['n']
                    x_2 = x_c + rad * np.cos(q_start + d_q)
                    # note sign on 2nd term is negative due to y sign convention (down positive)
                    y_2 = y_c - rad * np.sin(q_start + d_q)
                    nodes.append(np.array([len(nodes), x_2, y_2, 1, 1, 1, 1, 1.0]))
                # ------------------------
        else:
            # ------------------------
            # FLAT SECTION
            for j in range(0, g_1['n']):
                nodes.append(
                    np.array([
                        len(nodes), g_1['x'] + d_x * j / g_1['n'], g_1['y'] + d_y * j / g_1['n'], 1,
                        1, 1, 1, 1.0
                    ])
                )
            # ------------------------

    # GET THE LAST NODE ASSIGNED
    if rad == 0:
        if l_1 == 0 and l_2 == 0:
            i = 3
            nodes.append(np.array([len(nodes), geom[-1]['x'], geom[-1]['y'], 1, 1, 1, 1, 1.0]))
        else:
            i = 6
            nodes.append(np.array([len(nodes), geom[-1]['x'], geom[-1]['y'], 1, 1, 1, 1, 1.0]))
    else:
        if l_1 == 0 and l_2 == 0:
            i = 6
            nodes.append(np.array([len(nodes), geom[-1]['x'], geom[-1]['y'], 1, 1, 1, 1, 1.0]))
        else:
            i = 10
            nodes.append(np.array([len(nodes), geom[-1]['x'], geom[-1]['y'], 1, 1, 1, 1, 1.0]))

    for i in range(0, len(nodes) - 1):
        elements.append(np.array([i, i, i + 1, thick, 0]))

    return [np.array(nodes), np.array(elements)]


def template_out_to_in(sect):
    # BWS 2015
    # reference AISI Design Manual for the lovely corner radius calcs.
    # For template calc, convert outer dimensions and inside radii to centerline
    # dimensions throughout
    # convert the inner radii to centerline if nonzero
    if sect['type'] == 'C':
        b_1 = sect['b_1']
        b_2 = sect['b_2']
    else:
        b_1 = sect['b_1']
        b_2 = sect['b_2']

    thick = sect['t']
    if sect['r_out'] == 0:
        rad = 0
    else:
        rad = sect['r_out'] - thick/2
    depth = sect['d'] - thick/2 - rad - rad - thick/2

    if sect['l_1'] == 0:
        b_1 = b_1 - rad - thick/2
        l_1 = 0
    else:
        b_1 = b_1 - rad - thick/2 - (rad + thick/2) * np.tan(np.pi / 4)
        l_1 = sect['l_1'] - (rad + thick/2) * np.tan(np.pi / 4)
    if sect['l_2'] == 0:
        b_2 = b_2 - rad - thick/2
        l_2 = 0
    else:
        b_2 = b_2 - rad - thick/2 - (rad + thick/2) * np.tan(np.pi / 4)
        l_2 = sect['l_2'] - (rad + thick/2) * np.tan(np.pi / 4)
    return [depth, b_1, l_1, b_2, l_2, rad, thick]


def yield_mp(nodes, f_y, sect_props, restrained=False):
    # BWS
    # August 2000
    # [Py,Mxx_y,Mzz_y,M11_y,M22_y]
    # May 2019 trap nan when flat plate or other properites are zero

    f_yield = {'P': 0, 'Mxx': 0, 'Myy': 0, 'M11': 0, 'M22': 0}

    f_yield['P'] = f_y * sect_props['A']
    #account for the possibility of restrained bending vs. unrestrained bending
    if restrained is False:
        sect_props['Ixy'] = 0
    #Calculate stress at every point based on m_xx=1
    m_xx = 1
    m_yy = 0
    stress1 = np.zeros((1, len(nodes)))
    stress1 = stress1 - ((m_yy*sect_props['Ixx']
                        + m_xx*sect_props['Ixy'])
                       * (nodes[:, 1] - sect_props['cx'])
                       - (m_yy*sect_props['Ixy']
                          + m_xx*sect_props['Iyy'])
                       * (nodes[:, 2] - sect_props['cy'])) \
        / (sect_props['Iyy']*sect_props['Ixx'] - sect_props['Ixy']**2)
    if np.max(abs(stress1)) == 0:
        f_yield['Mxx'] = 0
    else:
        f_yield['Mxx'] = f_y / np.max(abs(stress1))
    #Calculate stress at every point based on m_yy=1
    m_xx = 0
    m_yy = 1
    stress1 = np.zeros((1, len(nodes)))
    stress1 = stress1 - ((m_yy*sect_props['Ixx']
                        + m_xx*sect_props['Ixy'])
                       * (nodes[:, 1] - sect_props['cx'])
                       - (m_yy*sect_props['Ixy']
                          + m_xx*sect_props['Iyy'])
                       * (nodes[:, 2] - sect_props['cy'])) \
        / (sect_props['Iyy']*sect_props['Ixx'] - sect_props['Ixy']**2)
    if np.max(abs(stress1)) == 0:
        f_yield['Myy'] = 0
    else:
        f_yield['Myy'] = f_y / np.max(abs(stress1))
    # %M11_y, M22_y
    # %transform coordinates of nodes into principal coordinates
    phi = sect_props['phi']
    transform = np.array([[np.cos(phi), -np.sin(phi)], [np.sin(phi), np.cos(phi)]])
    cent_coord = np.array([nodes[:, 1] - sect_props['cx'], nodes[:, 2] - sect_props['cy']])
    prin_coord = np.transpose(spla.inv(transform) @ cent_coord)
    f_yield['M11'] = 1
    stress1 = np.zeros((1, len(nodes)))
    stress1 = stress1 - f_yield['M11'] * prin_coord[:, 1] / sect_props['I11']
    if np.max(abs(stress1)) == 0:
        f_yield['M11'] = 0
    else:
        f_yield['M11'] = f_y / np.max(abs(stress1)) * f_yield['M11']

    f_yield['M22'] = 1
    stress1 = np.zeros((1, len(nodes)))
    stress1 = stress1 - f_yield['M22'] * prin_coord[:, 0] / sect_props['I22']
    if np.max(abs(stress1)) == 0:
        f_yield['M22'] = 0
    else:
        f_yield['M22'] = f_y / np.max(abs(stress1)) * f_yield['M22']
    return f_yield


def stress_gen(nodes, forces, sect_props, restrained=False, offset_basis=0):
    # BWS
    # 1998
    # offset_basis compensates for section properties that are based upon coordinate
    # [0, 0] being something other than the centreline of elements. For example,
    # if section properties are based upon the outer perimeter, then
    # offset_basis=[-thickness/2, -thickness/2]
    if isinstance(offset_basis, float) or isinstance(offset_basis, int):
        offset_basis = [offset_basis, offset_basis]

    stress = np.zeros((1, len(nodes)))
    stress = stress + forces['P'] / sect_props['A']
    if restrained:
        stress = stress - ((forces['Myy'] * sect_props['Ixx']) *
                           (nodes[:, 1] - sect_props['cx'] - offset_basis[0]) -
                           (forces['Mxx'] * sect_props['Iyy']) *
                           (nodes[:, 2] - sect_props['cy'] - offset_basis[1])
                           ) / (sect_props['Iyy'] * sect_props['Ixx'])
    else:
        stress = stress - ((forces['Myy'] * sect_props['Ixx'] + forces['Mxx'] * sect_props['Ixy']) *
                           (nodes[:, 1] - sect_props['cx'] - offset_basis[0]) -
                           (forces['Myy'] * sect_props['Ixy'] + forces['Mxx'] * sect_props['Iyy']) *
                           (nodes[:, 2] - sect_props['cy'] - offset_basis[1])
                           ) / (sect_props['Iyy'] * sect_props['Ixx'] - sect_props['Ixy']**2)
    phi = sect_props['phi'] * np.pi / 180
    transform = np.array([[np.cos(phi), -np.sin(phi)], [np.sin(phi), np.cos(phi)]])
    cent_coord = np.array([
        nodes[:, 1] - sect_props['cx'] - offset_basis[0],
        nodes[:, 2] - sect_props['cy'] - offset_basis[1]
    ])
    prin_coord = np.transpose(spla.inv(transform) @ cent_coord)
    stress = stress - \
        forces['M11'] * prin_coord[:, 1] / sect_props['I11']

    stress = stress - \
        forces['M22'] * prin_coord[:, 0] / sect_props['I22']
    nodes[:, 7] = stress.flatten()
    return nodes


def doubler(node, elem):
    # %BWS
    # %1998 (last modified)
    # %A function to double the number of elements to help
    # %out the discretization of the member somewhat.
    # %
    # %node=[node# x z dofx dofz dofy doftheta stress]
    # %elem=[elem# nodei nodej thickness]
    # %
    old_num_elem = len(elem)
    old_num_node = len(node)
    elem_out = np.zeros((2 * old_num_elem, 5))
    node_out = np.zeros((old_num_elem + old_num_node, 8))
    # %For node_out set all the old numbers to odd numbers and fill in the
    # %new ones with even numbers.
    for i in range(old_num_node):
        node_out[2 * i, 0] = 2 * node[i, 0]
        node_out[2 * i, 1:8] = node[i, 1:8]

    for i in range(old_num_elem):
        elem_out[2 * i, :] = [2 * elem[i, 0], 2 * elem[i, 1], 2*i + 1, elem[i, 3], elem[i, 4]]
        elem_out[2*i + 1, :] = [2*i + 1, 2*i + 1, 2 * elem[i, 2], elem[i, 3], elem[i, 4]]
        nnumi = int(elem[i, 1])
        nnumj = int(elem[i, 2])
        xcoord = np.mean([node[nnumi, 1], node[nnumj, 1]])
        zcoord = np.mean([node[nnumi, 2], node[nnumj, 2]])
        stress = np.mean([node[nnumi, 7], node[nnumj, 7]])
        node_out[2*i + 1, :] = [
            2*i + 1, xcoord, zcoord, node[nnumi, 3], node[nnumi, 4], node[nnumi, 5], node[nnumi, 6],
            stress
        ]
    return node_out, elem_out
